"""Matplotlib utilities which are used in
:py:class:`~dantro.plot.plot_helper.PlotHelper` or other places.
"""

import logging
from itertools import chain
from typing import List, Tuple

from .color_mngr import ColorManager

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


class figure_leak_prevention:
    """Context manager that aims to prevent superfluous matplotlib figures
    persisting beyond the context. Such figure objects can aggregate and start
    memory issues or even representation errors.

    Specifically, it does the following:

        * When entering, stores all current figure numbers
        * When exiting regularly, all figures that were opened within the
          context are closed, except the currently selected figure.
        * When exiting with an exception, the behaviour is the same, unless the
          ``close_current_fig_on_raise`` is set, in which case the currently
          selected figure is **not** excluded from closing.

    """

    def __init__(self, *, close_current_fig_on_raise: bool = False):
        """Initialize the context manager

        Args:
            close_current_fig_on_raise (bool, optional): If True, the
                currently selected figure will **not** be exempt from the
                figure closure in case an exception occurs. This flag has no
                effect when the context is exited without an exception.
        """
        self._fignums = None
        self._close_current = close_current_fig_on_raise

    def __enter__(self):
        """Upon entering, store all currently open figure numbers"""
        import matplotlib.pyplot as plt

        self._fignums = plt.get_fignums()
        log.trace(
            "Entering figure_leak_prevention context. Open figures: %s",
            self._fignums,
        )

    def __exit__(self, exc_type: type, *args) -> None:
        """Iterates over all currently open figures and closes all figures
        that were not previously open, except the currently selected figure.

        If an exception is detected, i.e. ``exc_type` is **not** None, the
        current figure is only closed if the context manager was entered with
        the ``close_current_fig_on_raise`` flag set.
        """
        import matplotlib.pyplot as plt

        log.trace(
            "Exiting figure_leak_prevention (exception: %s) ...", exc_type
        )

        # Determine whether to exclude the current figure or not
        exclude_current = not self._close_current or exc_type is None
        cfn = plt.gcf().number
        log.trace("  Current figure: %d", cfn)

        for n in plt.get_fignums():
            if n in self._fignums or (exclude_current and n == cfn):
                continue
            plt.close(n)
            log.trace("  Closed figure %d.", n)


def calculate_space_needed_hv(
    fig: "matplotlib.figure.Figure",
    obj,
    *,
    spacing_h: float = 0.02,
    spacing_v: float = 0.02,
) -> Tuple[float, float]:
    """Calculates the horizontal and vertical space needed for an object in
    this figure.

    .. note::

        This will invoke :py:meth:`matplotlib.figure.Figure.draw` two times
        and cause resizing of the figure!

    Args:
        fig (matplotlib.figure.Figure): The figure
        obj: The object in that figure to fit
        spacing_h (float, optional): Added to space needed (in inches)
        spacing_v (float, optional): Added to space needed (in inches)

    Returns:
        Tuple[float, float]: the horizontal and vertical space needed to fit
            into the figure (in inches).
    """

    def get_obj_width_height(obj) -> tuple:
        extent = obj.get_window_extent()
        return (extent.width / fig.dpi, extent.height / fig.dpi)

    # Draw the figure to have an object window
    fig.draw(fig.canvas.get_renderer())

    # Calculate and set the new width and height of the figure so the obj fits
    fig_width, fig_height = fig.get_figwidth(), fig.get_figheight()
    obj_width, obj_height = get_obj_width_height(obj)

    fig.set_figwidth(fig_width + obj_width)
    fig.set_figheight(fig_height + obj_height)

    # Draw again to get new object transformations
    fig.draw(fig.canvas.get_renderer())
    obj_width, obj_height = get_obj_width_height(obj)

    # Now calculate the needed space horizontally and vertically
    space_needed_h = obj_width / (fig_width + obj_width) + spacing_h
    space_needed_v = obj_height / (fig_height + obj_height) + spacing_v

    return space_needed_h, space_needed_v


def remove_duplicate_handles_labels(h: list, l: list) -> Tuple[list, list]:
    """Returns new aligned lists of handles and labels from which duplicates
    (identified by label) are removed.

    This maintains the order and association by keeping track of seen items;
    see https://stackoverflow.com/a/480227/1827608 for more information.

    Args:
        h (list): List of artist handles
        l (list): List of labels

    Returns:
        Tuple[list, list]: handles and labels
    """
    seen = set()
    hls = [
        (_h, _l) for _h, _l in zip(h, l) if not (_l in seen or seen.add(_l))
    ]
    h, l = [hl[0] for hl in hls], [hl[1] for hl in hls]
    return h, l


def gather_handles_labels(mpo) -> Tuple[list, list]:
    """Uses ``.findobj`` to search a figure or axis for legend objects and
    returns lists of handles and (string) labels.
    """
    import matplotlib as mpl

    h, l = [], []
    for lg in mpo.findobj(mpl.legend.Legend):
        h += [_h for _h in lg.legend_handles]
        l += [_t.get_text() for _t in lg.texts]

    # Remove duplicates and return
    return remove_duplicate_handles_labels(h, l)


def prepare_legend_args(
    h, l, *, custom_labels: List[str], hiding_threshold: int
) -> Tuple[list, list, bool]:
    """A utility function that allows setting custom legend handles and
    implements some logic to hide all handles if there are too many."""
    import matplotlib as mpl

    # Might want to use custom labels
    if custom_labels:
        log.remark("Using custom labels:  " + ", ".join(custom_labels))
        l = custom_labels

    # Evaluate the hiding threshold
    past_thresh = (
        hiding_threshold is not None and min(len(h), len(l)) > hiding_threshold
    )
    if past_thresh:
        log.remark(
            "With %d handles and %d labels, passed hiding threshold of %d.",
            len(h),
            len(l),
            hiding_threshold,
        )
    else:
        log.remark(
            "Have %d handles and %d labels available for the legend.",
            len(h),
            len(l),
        )

    return h, l, past_thresh


def set_tick_locators_or_formatters(
    *,
    ax: "matplotlib.axes.Axes",
    kind: str,
    x: dict = None,
    y: dict = None,
    z: dict = None,
):
    """Sets the tick locators or formatters.
    Look at the :py:class:`~dantro.plot.plot_helper.PlotHelper` methods
    ``_hlpr_set_tick_{locators/formatters}`` for more information.

    Names are looked up in the :py:mod:`matplotlib.ticker` and
    :py:mod:`matplotlib.dates` modules.

    Args:
        ax (matplotlib.axes.Axes): The axes object
        kind (str): Whether to set a ``locator`` or a ``formatter``.
        x (dict, optional): The config for the x-axis tick locator/formatter
        y (dict, optional): The config for the y-axis tick locator/formatter
        z (dict, optional): The config for the z-axis tick locator/formatter
    """
    # Safe guard against calling this with unexpected arguments from
    # within the actual helper methods; not part of public interface.
    if kind not in ("locator", "formatter"):
        raise ValueError(f"Bad kind: {kind}")

    def _set_locator_or_formatter(
        *,
        _ax: "matplotlib.axes.Axes",
        _axis: str,
        _major: bool,
        _kind: str,
        name: str,
        args: tuple = (),
        **kwargs,
    ):
        """Set the tick locator or formatter on a specific axis.

        Args:
            _ax (matplotlib.axes.Axes): The axes object to work on.
            _axis (str):    The axis name, ``x`` or ``y``
            _major (bool):  Whether to set the major or minor ticks
            _kind (str):    The kind of function to set: ``locator`` or
                ``formatter``
            name (str):     The name of the locator or formatter, looked up in
                :py:mod:`matplotlib.ticker` and :py:mod:`matplotlib.dates`.
            args (tuple):   Args passed on to the respective locator
                            or formatter setter function.
            **kwargs:       Kwargs passed on to the respective locator
                            or formatter setter function.
        """
        import matplotlib as mpl

        # Get the ticker from the name, looking it up in both the mpl.ticker
        # and the mpl.dates modules
        try:
            ticker = getattr(mpl.ticker, name)
        except AttributeError as err:
            try:
                ticker = getattr(mpl.dates, name)
            except AttributeError as err:
                # Customize the error message for (i) no name (ii) wrong name
                # for locators and formatters.
                _avail = ", ".join(
                    s
                    for s in chain(dir(mpl.ticker), dir(mpl.dates))
                    if _kind.capitalize() in s
                )
                raise AttributeError(
                    f"The given {_kind} name '{name}' is not valid! "
                    f"Choose from: {_avail}"
                ) from err

        # Get the locator or formatter function for the respective
        # major or minor axis.
        ax_obj = getattr(_ax, f"{_axis}axis")
        setter = getattr(
            ax_obj, f"set_{'major' if _major else 'minor'}_{_kind}"
        )

        try:
            setter(ticker(*args, **kwargs))
        except Exception as exc:
            raise ValueError(
                f"Failed setting {'major' if _major else 'minor'} {_kind} "
                f"'{name}' for {_axis}-axis! Check the matplotlib "
                "documentation for valid arguments. "
                f"Got:\n  args: {args}\n  kwargs: {kwargs}"
            ) from exc

    # Decide which tick locator or formatter to set, and set it
    if x:
        if x.get("major"):
            _set_locator_or_formatter(
                _ax=ax,
                _kind=kind,
                _axis="x",
                _major=True,
                **x["major"],
            )

        if x.get("minor"):
            _set_locator_or_formatter(
                _ax=ax,
                _kind=kind,
                _axis="x",
                _major=False,
                **x["minor"],
            )

    if y:
        if y.get("major"):
            _set_locator_or_formatter(
                _ax=ax,
                _kind=kind,
                _axis="y",
                _major=True,
                **y["major"],
            )

        if y.get("minor"):
            _set_locator_or_formatter(
                _ax=ax,
                _kind=kind,
                _axis="y",
                _major=False,
                **y["minor"],
            )

    if hasattr(ax, "zaxis"):
        if z:
            if z.get("major"):
                _set_locator_or_formatter(
                    _ax=ax,
                    _kind=kind,
                    _axis="z",
                    _major=True,
                    **z["major"],
                )

            if z.get("minor"):
                _set_locator_or_formatter(
                    _ax=ax,
                    _kind=kind,
                    _axis="z",
                    _major=False,
                    **z["minor"],
                )



"""A module that implements a bunch of plot utilities used in the plotting
functions. These can be shared tools between the plotting functions.
"""

import copy

import numpy as np

# -----------------------------------------------------------------------------


def plot_errorbar(
    *,
    ax,
    x: np.ndarray,
    y: np.ndarray,
    yerr: np.ndarray,
    fill_between: bool = False,
    fill_between_kwargs: dict = None,
    **errorbar_kwargs,
):
    """Given the data and (optionally) the y-error data, plots a single
    errorbar line. With ``fill_between=True``, a shaded area is plotted instead
    of the errorbar markers.

    The following ``fill_between_kwargs`` defaults are assumed:

        - ``color = line_color``
        - ``alpha = 0.2 * line_alpha``
        - ``lw = 0.``

    Args:
        ax: The axis to plot on
        x (numpy.ndarray): The x data to use
        y (numpy.ndarray): The y-data to use for ``ax.errorbar``. Needs to be
            1D and have coordinates associated which will be used for the
            x-values.
        yerr (numpy.ndarray): The y-error data
        fill_between (bool, optional): Whether to use plt.fill_between or
            plt.errorbar to plot y-errors
        fill_between_kwargs (dict, optional): Passed on to plt.fill_between
        **errorbar_kwargs: Passed on to plt.errorbar

    Raises:
        ValueError: On non-1D data

    Returns:
        The matplotlib legend handle of the errorbar line or of the errorbands
    """
    if y.ndim != 1 or (yerr is not None and yerr.ndim != 1):
        raise ValueError(
            "Expected 1D `y` and `yerr` data for errorbar plot but selected "
            f"data was {y.ndim}- and {yerr.ndim}-dimensional, respecetively!\n"
            f"\ny: {y}\nyerr: {yerr}"
        )

    # Data is ok.
    # Plot the data against its coordinates, including the y-error data only if
    # no fill-between is to be performed
    ebar = ax.errorbar(
        x, y, yerr=yerr if not fill_between else None, **errorbar_kwargs
    )

    if not fill_between or yerr is None:
        # Return the regular errorbar artist
        return ebar
    # else: plot yerr as shaded area using fill_between

    # Find out the colour of the error bar line by inspecting line collection
    lc, _, _ = ebar
    line_color = lc.get_c()
    line_alpha = lc.get_alpha() if lc.get_alpha() else 1.0

    # Prepare fill between arguments, setting some default values
    fb_kwargs = (
        copy.deepcopy(fill_between_kwargs) if fill_between_kwargs else {}
    )

    fb_kwargs["color"] = fb_kwargs.get("color", line_color)
    fb_kwargs["alpha"] = fb_kwargs.get("alpha", line_alpha * 0.2)
    fb_kwargs["lw"] = fb_kwargs.get("lw", 0.0)

    # Fill.
    fb = ax.fill_between(x, y1=(y - yerr), y2=(y + yerr), **fb_kwargs)

    # Return the artist that is to be used in the legend
    return (ebar, fb)



""":py:mod:`dantro` provides a uniform interface for hierarchically structured
and semantically heterogeneous data.
It is built around three main features:

- **data handling**: loading heterogeneous data into a tree-like data
  structure, providing a uniform interface to it
- **data transformation**: performing arbitrary operations on the data, if
  necessary using lazy evaluation
- **data visualization**: creating a visual representation of the processed
  data

Together, these stages constitute a **data processing pipeline**:
an automated sequence of predefined, configurable operations.

See :ref:`the user manual <welcome>` for more information.
"""

__version__ = "0.19.2"
"""Package version"""

# Set up the root logger such that the logging configuration is applied
from .logging import getLogger as _getLogger

_log = _getLogger(__name__)

# -- Most important dantro classes --------------------------------------------
from .data_mngr import DataManager
from .plot_mngr import PlotManager



"""Implements registration of data loaders, including a decorator to ensure
correct loader function signature (which also automatically keeps track of the
data loader function).
"""

import logging
from typing import Callable, List

from .._registry import ObjectRegistry as _ObjectRegistry

log = logging.getLogger(__name__)

LOAD_FUNC_PREFIX: str = "_load_"
"""The prefix that all load functions need to start with"""

# -----------------------------------------------------------------------------


class DataLoaderRegistry(_ObjectRegistry):
    """Specialization of :py:class:`~dantro._registry.ObjectRegistry` for the
    purpose of keeping track of data loaders.
    """

    _DESC = "data loader"
    _SKIP = False
    _OVERWRITE = False
    _EXPECTED_TYPE = None


DATA_LOADERS = DataLoaderRegistry()
"""The dantro data loaders registry.

The :py:class:`~dantro.data_mngr.DataManager` and derived classes have access
to all data loaders via this registry (in addition to method-based access
they have via potentially used mixins).

To register a new loader, use the :py:func:`.add_loader` decorator:
"""


def _register_loader(
    wrapped_func: Callable,
    name: str,
    *,
    skip_existing: bool = False,
    overwrite_existing: bool = True,
) -> None:
    """Internally used method to add an entry to the shared loader registry.

    Args:
        wrapped_func (Callable): The wrapped callable that is to be registered
            as a loader. This is what the :py:func:`.add_loader` decorator
            generates.
        name (str, optional): The name to use for registration.
        skip_existing (bool, optional): Whether to skip registration if the
            loader name is already registered. This suppresses the
            ValueError raised on existing loader name.
        overwrite_existing (bool, optional): Whether to overwrite a potentially
            already existing loader of the same name. If set, this takes
            precedence over ``skip_existing``.
    """
    return DATA_LOADERS.register(
        wrapped_func,
        name=name,
        skip_existing=skip_existing,
        overwrite_existing=overwrite_existing,
    )


def add_loader(
    *,
    TargetCls: type,
    omit_self: bool = True,
    overwrite_existing: bool = True,
    register_aliases: List[str] = None,
):
    """This decorator should be used to specify loader methods in mixin classes
    to the :py:class:`~dantro.data_mngr.DataManager`.

    All decorated methods where ``omit_self is True`` will additinoally be
    registered in the :py:data:`.DATA_LOADERS` registry.

    Example:

    .. testcode::

        from dantro.containers import ObjectContainer
        from dantro.data_loaders import add_loader

        class MyDataLoaderMixin:

            @add_loader(TargetCls=ObjectContainer)
            def _load_foobar(path: str, *, TargetCls: type, **kws):
                # load something from the given file path
                with open(path, **kws) as f:
                    data = f.read()

                return TargetCls(data=data)

        # Define a DataManager that has the custom loader mixed-in

        from dantro import DataManager

        class MyDataManager(MyDataLoaderMixin, DataManager):
            pass

    .. testcode::
        :hide:

        from dantro.data_loaders._registry import DATA_LOADERS

        assert "foobar" in DATA_LOADERS
        del DATA_LOADERS._d["foobar"]
        assert "foobar" not in DATA_LOADERS

    .. note::

        Loader methods need to be named ``_load_<name>`` and are then
        accessible via ``<name>``.

        **Important:** Loader methods may not be named ``_load_file``!

    .. hint::

        This decorator **can also be used on standalone functions**, without
        the need to define a mixin class.
        In such a case, ``omit_self`` can still be set to False, leading to the
        first positional argument that the decorated function needs to accept
        to be the :py:class:`~dantro.data_mngr.DataManager` instance that the
        loader is used in.

        Note that these standalone function should still begin with ``_load_``.

    Args:
        TargetCls (type): The return type of the load function. This is stored
            as an attribute of the decorated function.
        omit_self (bool, optional): If True (default), the decorated method
            will not be supplied with the ``self`` object instance, thus being
            equivalent to a class method.
        overwrite_existing (bool, optional): If False, will not overwrite the
            existing registry entry in :py:data:`.DATA_LOADERS` but raise an
            error instead.
        register_aliases (List[str], optional): If given, will additionally
            register this method under the given name
    """

    def load_func_decorator(func):
        """This decorator sets the load function's ``TargetCls`` attribute."""

        func_name = func.__name__
        assert func_name.startswith(LOAD_FUNC_PREFIX)
        assert func_name != "_load_file"  # used in DataManager

        name = func_name[len(LOAD_FUNC_PREFIX) :]

        # Wrap the load function such that the call signature is handled
        def load_func(dm: "DataManager", *args, **kwargs):
            """Calls the load function, either with or without ``self``."""
            load_func.__doc__ = func.__doc__
            if omit_self:
                # class method or standalone function
                return func(*args, **kwargs)
            # regular method
            return func(dm, *args, **kwargs)

        # Set the target class as function attribute and also make the captured
        # function more easily available
        load_func.TargetCls = TargetCls
        load_func._func = func

        # Carry over the docstring of the decorated function
        load_func.__doc__ = func.__doc__

        # Keep track of it via the registry dict.
        # NOTE Can only do this for class methods, which can act as standalone
        #      functions. Load functions that require access to other
        #      attributes or methods of the DataManager *object* can not be
        #      registered in a sensible way because these methods may not be
        #      available.
        if omit_self:
            _register_loader(
                load_func, name=name, overwrite_existing=overwrite_existing
            )
            for alias in register_aliases if register_aliases else ():
                _register_loader(
                    load_func,
                    name=alias,
                    overwrite_existing=overwrite_existing,
                )

        return load_func

    return load_func_decorator



"""Implements config-configurable function invocation that can be used for
applying function calls to a plot.
This is used in :py:func:`~dantro.plot.funcs.multiplot.multiplot` plot and
the :py:meth:`~dantro.plot.plot_helper.PlotHelper._hlpr_call` helper function.
"""

import copy
import logging
from typing import Any, Callable, Dict, List, Tuple, Union

from ..._import_tools import (
    LazyLoader,
    import_module_or_object,
    resolve_lazy_imports,
)
from ...exceptions import *
from ...tools import make_columns, recursive_update
from ..utils.color_mngr import ColorManager, parse_cmap_and_norm_kwargs

log = logging.getLogger(__name__)

# Lazy module loading for packages that take a long time to import
_plt = LazyLoader("matplotlib.pyplot", _depth=1)
_sns = LazyLoader("seaborn", _depth=1)
# NOTE This import is specifically for the definition of the seaborn-based
#      multiplot functions and should not be used for anything else!
#      Depth 1 means: `_sns.foobar` is still a LazyLoader object and will only
#      be resolved upon an *attribute* call! Subsequently, these need to be
#      resolved manually, as done via `resolve_lazy_imports`.

# fmt: off

# -----------------------------------------------------------------------------

# Define some default plot functions
# NOTE These are defined here in a lazy fashion, thus being actually
#      LazyLoader instances. They are (and need to be) resolved upon a first
#      call to the multiplot function
MULTIPLOT_FUNC_KINDS = { # --- start literalinclude
    # Seaborn - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # https://seaborn.pydata.org/api.html

    # Relational plots
    "sns.scatterplot":      _sns.scatterplot,
    "sns.lineplot":         _sns.lineplot,

    # Distribution plots
    "sns.histplot":         _sns.histplot,
    "sns.kdeplot":          _sns.kdeplot,
    "sns.ecdfplot":         _sns.ecdfplot,
    "sns.rugplot":          _sns.rugplot,

    # Categorical plots
    "sns.stripplot":        _sns.stripplot,
    "sns.swarmplot":        _sns.swarmplot,
    "sns.boxplot":          _sns.boxplot,
    "sns.violinplot":       _sns.violinplot,
    "sns.boxenplot":        _sns.boxenplot,
    "sns.pointplot":        _sns.pointplot,
    "sns.barplot":          _sns.barplot,
    "sns.countplot":        _sns.countplot,

    # Regression plots
    "sns.regplot":          _sns.regplot,
    "sns.residplot":        _sns.residplot,

    # Matrix plots
    "sns.heatmap":          _sns.heatmap,

    # Utility functions
    "sns.despine":          _sns.despine,

    # Matplotlib - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # https://matplotlib.org/tutorials/introductory/sample_plots.html

    # Relational plots
    "plt.fill":             _plt.fill,
    "plt.scatter":          _plt.scatter,
    "plt.plot":             _plt.plot,
    "plt.polar":            _plt.polar,
    "plt.loglog":           _plt.loglog,
    "plt.semilogx":         _plt.fill,
    "plt.semilogy":         _plt.semilogy,
    "plt.errorbar":         _plt.errorbar,

    # Distribution plots
    "plt.hist":             _plt.hist,
    "plt.hist2d":           _plt.hist2d,

    # Categorical plots
    "plt.bar":              _plt.bar,
    "plt.barh":             _plt.barh,
    "plt.pie":              _plt.pie,
    "plt.table":            _plt.table,

    # Matrix plots
    "plt.imshow":           _plt.imshow,
    "plt.pcolormesh":       _plt.pcolormesh,

    # Vector plots
    "plt.contour":          _plt.contour,
    "plt.quiver":           _plt.quiver,
    "plt.streamplot":       _plt.streamplot,
}   # --- end literalinclude
"""The default-available plot kinds for the
:py:func:`~dantro.plot.funcs.multiplot.multiplot` function.

Details of the seaborn-related plots can be found here in the
`seaborn docs <https://seaborn.pydata.org/api.html>`_.
"""

MULTIPLOT_CAUTION_FUNC_NAMES = tuple(
    func_name for func_name in MULTIPLOT_FUNC_KINDS
    if func_name not in (
        "sns.despine",
    )
)
"""The multiplot functions that emit a warning if they do not get any arguments
when called. This is helpful for functions that e.g. require a ``data``
argument but do *not* fail or warn if no such argument is passed on to them.
"""

# fmt: on

# -----------------------------------------------------------------------------


def parse_function_specs(
    *,
    _hlpr: "PlotHelper",
    _funcs: Dict[str, Callable] = None,
    _shared_kwargs: dict = {},
    function: Union[str, Callable, Tuple[str, str]],
    args: list = None,
    pass_axis_object_as: str = None,
    pass_helper: bool = False,
    **func_kwargs,
) -> Tuple[str, Callable, list, dict]:
    """Parses a function specification used in the ``invoke_function`` helper.
    If ``function`` is a string it is looked up from the ``_funcs`` dict.

    See :py:func:`~.parse_and_invoke_function` and
    :py:func:`~dantro.plot.funcs.multiplot.multiplot`.

    Args:
        _hlpr (dantro.plot.plot_helper.PlotHelper): The currently
            used PlotHelper instance
        _funcs (Dict[str, Callable]): The lookup dictionary for callables
        _shared_kwargs (dict, optional): Shared kwargs that passed on to
            all multiplot functions. They are recursively updated with
            the individual plot functions' ``func_kwargs``.
        function (Union[str, Callable, Tuple[str, str]]): The callable
            function object or the name of the plot function to look up.
            If given as 2-tuple ``(module, name)``, will attempt an import of
            that module.
        args (list, optional): The positional arguments for the plot function
        pass_axis_object_as (str, optional): If given, will add a keyword
            argument with this name to pass the current axis object to the
            to-be-invoked function.
        pass_helper (bool, optional): If true, passes the helper instance to
            the function call as keyword argument ``hlpr``.
        **func_kwargs (dict): The function kwargs to be passed on to the
            function object.

    Returns:
        Tuple[str, Callable, list, dict]: A tuple of function name, callable,
            positional arguments, and keyword arguments.
    """
    # Parse positional and keyword arguments
    if args is None:
        args = []

    if _funcs is None:
        _funcs = MULTIPLOT_FUNC_KINDS

    func_kwargs = recursive_update(copy.deepcopy(_shared_kwargs), func_kwargs)
    func_kwargs = parse_cmap_and_norm_kwargs(**func_kwargs)

    if pass_axis_object_as:
        func_kwargs[pass_axis_object_as] = _hlpr.ax

    if pass_helper:
        func_kwargs["hlpr"] = _hlpr

    # Get the function object and a readable name
    if callable(function):
        func_name = function.__name__
        func = function

    elif isinstance(function, (list, tuple)):
        # Import
        mod, name = function
        func = import_module_or_object(mod, name)
        func_name = ".".join(function)

    else:
        # Look up the function in the `_funcs` dict.
        # Still need to resolve all lazy imports in the lookup dictionary.
        resolve_lazy_imports(_funcs)

        func_name = function
        try:
            func = _funcs[func_name]

        except KeyError as err:
            _avail = " (none)\n"
            if _funcs:
                _avail = make_columns(_funcs)

            raise ValueError(
                f"A function called '{func_name}' could not be found "
                f"by name!\nAvailable functions:\n{_avail}\n"
                "Alternatively, pass a callable instead of the function name "
                "or pass a 2-tuple of (module, name) to import a callable."
            ) from err

    return func_name, func, args, func_kwargs


def parse_and_invoke_function(
    *,
    hlpr: "PlotHelper",
    shared_kwargs: dict,
    func_kwargs: dict,
    show_hints: bool,
    call_num: int,
    funcs: Dict[str, Callable] = None,
    caution_func_names: List[str] = None,
) -> Any:
    """Parses function arguments and then calls
    :py:func:`~dantro.plot.funcs.multiplot.multiplot`.

    Args:
        hlpr (PlotHelper): The currently used PlotHelper instance
        funcs (Dict[str, Callable], optional): The lookup dictionary for the
            plot functions. If not given, will use a default lookup dictionary
            with a set of seaborn and matplotlib functions.
        shared_kwargs (dict): Arguments shared between function calls
        func_kwargs (dict): Arguments for *this* function in particular
        show_hints (bool): Whether to show hints
        call_num (int): The number of this plot, for easier identification
        caution_func_names (List[str], optional): a list of function names that
            will trigger a log message if no function kwargs were given.
            If not explicitly given, will use some defaults.

    Returns:
        Any: return value of plot function call
    """
    if caution_func_names is None:
        caution_func_names = MULTIPLOT_CAUTION_FUNC_NAMES

    # Get the function name, the function object and all function kwargs
    # from the configuration entry.
    func_name, func, func_args, func_kwargs = parse_function_specs(
        _hlpr=hlpr, _funcs=funcs, _shared_kwargs=shared_kwargs, **func_kwargs
    )

    # Notify user if plot functions do not get any kwargs passed on.
    # This is e.g. helpful and relevant for seaborn functions that require
    # a 'data' kwarg but do not fail or warn if no 'data' is passed on to them.
    if show_hints and not func_kwargs and func_name in caution_func_names:
        log.caution(
            "You seem to have called '%s' without any function arguments. "
            "If this produces unexpected output, check that all required "
            "arguments (e.g. `data`, `x`, ...) were given.\n"
            "To silence this warning, set `show_hints` to `False`.",
            func_name,
        )

    # Apply the plot function and allow it to fail to make sure that potential
    # other plots are still plotted and shown.
    rv = None
    try:
        rv = func(*func_args, **func_kwargs)

    except Exception as exc:
        msg = (
            f"The call to '{func_name}' (call no. {call_num} on axis "
            f"{hlpr.ax_coords}) did not succeed!\n"
            f"Got a {type(exc).__name__}: {exc}"
        )
        if hlpr.raise_on_error:
            raise PlottingError(msg) from exc
        log.warning(
            f"{msg}\nEnable debug mode to get a full traceback. "
            "Proceeding with next plot ..."
        )

    return rv



"""Implements loading of Hdf5 files into the dantro data tree"""

import logging
import os
from typing import Dict, Union

import numpy as np

from .._import_tools import LazyLoader
from ..base import BaseDataContainer, BaseDataGroup
from ..containers import NumpyDataContainer
from ..groups import OrderedDataGroup
from ..proxy import Hdf5DataProxy
from ..tools import decode_bytestrings, print_line
from ._registry import add_loader

log = logging.getLogger(__name__)

h5 = LazyLoader("h5py")

# -----------------------------------------------------------------------------


class Hdf5LoaderMixin:
    """Supplies functionality to load HDF5 files into the
    :py:class:`~dantro.data_mngr.DataManager`.

    It resolves the HDF5 groups into corresponding data groups and the datasets
    (by default) into
    :py:class:`~dantro.containers.numeric.NumpyDataContainer` s.

    If ``enable_mapping`` is set, the class variables ``_HDF5_DSET_MAP`` and
    ``_HDF5_GROUP_MAP`` are used to map from a string to a container type. The
    class variable ``_HDF5_MAP_FROM_ATTR`` determines the default value of the
    attribute to read and use as input string for the mapping.
    """

    _HDF5_DSET_DEFAULT_CLS: type = NumpyDataContainer
    """the default class to use for datasets. This should be a dantro
    :py:class:`~dantro.base.BaseDataContainer`-derived class.
    Note that certain data groups can overwrite the default class for
    underlying members."""

    _HDF5_GROUP_MAP: Dict[str, type] = None
    """If mapping is enabled, the equivalent dantro types for HDF5 groups are
    determined from this mapping."""

    _HDF5_DSET_MAP: Dict[str, type] = None
    """If mapping is enabled, the equivalent dantro types for HDF5 datasets
    are determined from this mapping."""

    _HDF5_MAP_FROM_ATTR: str = None
    """The name of the HDF5 dataset or group attribute to read in order to
    determine the type mapping. For example, this could be ``"content"``.
    This is the fallback value if no ``map_from_attr`` argument is given to
    :py:meth:`dantro.data_loaders.hdf5.Hdf5LoaderMixin._load_hdf5`"""

    _HDF5_DECODE_ATTR_BYTESTRINGS: bool = True
    """If true (default), will attempt to decode HDF5 attributes that are
    stored as byte arrays into regular Python strings; this can make attribute
    handling much easier."""

    @add_loader(TargetCls=OrderedDataGroup, omit_self=False)
    def _load_hdf5(
        self,
        filepath: str,
        *,
        TargetCls: type,
        load_as_proxy: bool = False,
        proxy_kwargs: dict = None,
        lower_case_keys: bool = False,
        enable_mapping: bool = False,
        map_from_attr: str = None,
        direct_insertion: bool = False,
        progress_params: dict = None,
    ) -> OrderedDataGroup:
        """Loads the specified hdf5 file into DataGroup- and DataContainer-like
        objects; this completely recreates the hierarchic structure of the hdf5
        file. The data can be loaded into memory completely, or be loaded as
        a proxy object.

        The :py:class:`h5py.File` and :py:class:`h5py.Group` objects will be
        converted to the specified
        :py:class:`~dantro.base.BaseDataGroup`-derived objects and the
        :py:class:`h5py.Dataset` objects to the specified
        :py:class:`~dantro.base.BaseDataContainer`-derived object.

        All HDF5 group or dataset attributes are carried over and are
        accessible under the ``attrs`` attribute of the respective dantro
        objects in the tree.

        Args:
            filepath (str): The path to the HDF5 file that is to be loaded
            TargetCls (type): The group type this is loaded into
            load_as_proxy (bool, optional): if True, the leaf datasets are
                loaded as :py:class:`dantro.proxy.hdf5.Hdf5DataProxy`
                objects. That way, the data is only loaded into memory when
                their ``.data`` property is accessed the first time, either
                directly or indirectly.
            proxy_kwargs (dict, optional): When loading as proxy, these
                parameters are unpacked in the ``__init__`` call. For available
                argument see :py:class:`~dantro.proxy.hdf5.Hdf5DataProxy`.
            lower_case_keys (bool, optional): whether to use only lower-case
                versions of the paths encountered in the HDF5 file.
            enable_mapping (bool, optional): If true, will use the class
                variables ``_HDF5_GROUP_MAP`` and ``_HDF5_DSET_MAP`` to map
                groups or datasets to a custom container class during loading.
                Which attribute to read is determined by the ``map_from_attr``
                argument (see there).
            map_from_attr (str, optional): From which attribute to read the
                key that is used in the mapping. If nothing is given, the
                class variable ``_HDF5_MAP_FROM_ATTR`` is used.
            direct_insertion (bool, optional): If True, some non-crucial checks
                are skipped during insertion and elements are inserted (more
                or less) directly into the data tree, thus speeding up the data
                loading process.
                This option should only be enabled if data is loaded into a yet
                unpopulated part of the data tree, otherwise existing elements
                might be overwritten silently.
                This option only applies to data groups, not to containers.
            progress_params (dict, optional): parameters for the progress
                indicator. Possible keys:

                level (int):
                    how verbose to print progress info; possible values are:
                    ``0``: None, ``1``: on file level, ``2``: on dataset level.
                    Note that this option and the ``progress_indicator`` of
                    the DataManager are independent from each other.
                fstr:
                    format string for progress report, receives the following
                    keys:

                        * ``progress_info`` (total progress indicator),
                        * ``fname`` (basename of current hdf5 file),
                        * ``fpath`` (full path of current hdf5 file),
                        * ``name`` (current dataset name),
                        * ``path`` (current path within the hdf5 file)

        Returns:
            OrderedDataGroup: The populated root-level group, corresponding to
                the base group of the file

        Raises:
            ValueError: If ``enable_mapping``, but no map attribute can be
                determined from the given argument or the class variable
                ``_HDF5_MAP_FROM_ATTR``
        """
        # Initialize the root group
        log.debug(
            "Loading hdf5 file %s into %s ...", filepath, TargetCls.__name__
        )
        root = TargetCls()

        # Get the classes to use for groups and/or containers
        DsetCls = self._HDF5_DSET_DEFAULT_CLS

        # Determine from which attribute to read the mapping
        if not map_from_attr:
            # No custom value was given; use the class variable, if available
            if self._HDF5_MAP_FROM_ATTR:
                map_from_attr = self._HDF5_MAP_FROM_ATTR

            elif enable_mapping:
                # Mapping was enabled but it is unclear from which attribute
                # the map should be read. Need to raise an exception
                raise ValueError(
                    "Could not determine from which attribute to read the "
                    "mapping. Either set the loader argument `map_from_attr`, "
                    "the class variable _HDF5_MAP_FROM_ATTR, or disable "
                    "mapping altogether via the `enable_mapping` argument."
                )

        # Prepare parameters
        GroupMap = self._HDF5_GROUP_MAP if enable_mapping else {}
        DsetMap = self._HDF5_DSET_MAP if enable_mapping else {}

        # Prepare progress information
        progress_params = progress_params if progress_params else {}
        plvl = progress_params.get("level", 0)
        pfstr = progress_params.get(
            "fstr", "  {progress_info:} {fname:} : {path:} ... "
        )

        # Now recursively go through the hdf5 file and add them to the roo
        with h5.File(filepath, "r") as h5file:
            if plvl >= 1:
                # Print information on the level of this file
                _info = pfstr.format(
                    progress_info=self._progress_info_str,
                    fpath=filepath,
                    fname=os.path.basename(filepath),
                    key="",
                    path="",
                )
                print_line(_info)

            # Load the file level attributes, manually re-creating the dict
            root.attrs = {
                k: self._decode_attr_val(v) for k, v in h5file.attrs.items()
            }

            # Now recursively load the data into the root group
            self._recursively_load_hdf5(
                h5file,
                target=root,
                load_as_proxy=load_as_proxy,
                proxy_kwargs=proxy_kwargs,
                lower_case_keys=lower_case_keys,
                DsetCls=DsetCls,
                GroupMap=GroupMap,
                DsetMap=DsetMap,
                map_attr=map_from_attr,
                direct_insertion=direct_insertion,
                plvl=plvl,
                pfstr=pfstr,
            )

        return root

    @add_loader(TargetCls=OrderedDataGroup, omit_self=False)
    def _load_hdf5_proxy(self, *args, **kwargs) -> OrderedDataGroup:
        """This is a shorthand for
        :py:meth:`~dantro.data_loaders.hdf5.Hdf5LoaderMixin._load_hdf5`
        with the ``load_as_proxy`` flag set.
        """
        return self._load_hdf5(*args, load_as_proxy=True, **kwargs)

    @add_loader(TargetCls=OrderedDataGroup, omit_self=False)
    def _load_hdf5_as_dask(self, *args, **kwargs) -> OrderedDataGroup:
        """This is a shorthand for
        :py:meth:`~dantro.data_loaders.hdf5.Hdf5LoaderMixin._load_hdf5`
        with the ``load_as_proxy`` flag set and ``resolve_as_dask`` passed as
        additional arguments to the proxy via ``proxy_kwargs``.
        """
        return self._load_hdf5(
            *args,
            load_as_proxy=True,
            proxy_kwargs=dict(resolve_as_dask=True),
            **kwargs,
        )

    # .........................................................................

    def _recursively_load_hdf5(
        self,
        src: Union["h5py.Group", "h5py.File"],
        *,
        target: BaseDataGroup,
        lower_case_keys: bool,
        direct_insertion: bool,
        **kwargs,
    ):
        """Recursively loads the data from a source object (an h5py.File or a
        h5py.Group) into the target dantro group.

        Args:
            src (Union[h5py.Group, h5py.File]): The HDF5 source object from
                which to load the data. This object it iterated over.
            target (BaseDataGroup): The target group to populate with the data
                from ``src``.
            lower_case_keys (bool): Whether to make keys lower-case
            direct_insertion (bool): Whether to use direct insertion mode on
                the target group (and all groups below)
            **kwargs: Passed on to the group and container loader methods,
                :py:meth:`~dantro.data_loaders.hdf5.Hdf5LoaderMixin._container_from_h5dataset`
                and
                :py:meth:`~dantro.data_loaders.hdf5.Hdf5LoaderMixin._group_from_h5group`.

        Raises:
            NotImplementedError: When encountering objects other than groups
                or datasets in the HDF5 file
        """
        # Go through the elements of the source object
        for key, obj in src.items():
            if lower_case_keys and isinstance(key, str):
                key = key.lower()

            with target._direct_insertion_mode(enabled=direct_insertion):
                if isinstance(obj, h5.Group):
                    # Create the new group
                    grp = self._group_from_h5group(
                        obj, target=target, name=key, **kwargs
                    )

                    # Continue recursion
                    self._recursively_load_hdf5(
                        obj,
                        target=grp,
                        lower_case_keys=lower_case_keys,
                        direct_insertion=direct_insertion,
                        **kwargs,
                    )

                elif isinstance(obj, h5.Dataset):
                    # Reached a leaf -> Import the data and attributes into a
                    # BaseDataContainer-derived object. This assumes that the
                    # given DsetCls supports np.ndarray-like data
                    self._container_from_h5dataset(
                        obj, target=target, name=key, **kwargs
                    )

                else:
                    raise NotImplementedError(
                        f"Object {key} is neither a dataset nor a group, but "
                        f"of type {type(obj)}. Cannot load this!"
                    )

    def _group_from_h5group(
        self,
        h5grp: "h5py.Group",
        target: BaseDataGroup,
        *,
        name: str,
        map_attr: str,
        GroupMap: dict,
        **_,
    ) -> BaseDataGroup:
        """Adds a new group from a h5.Group

        The group types may be mapped to different dantro types; this is
        controlled by the extracted HDF5 attribute with the name specified in
        the ``_HDF5_MAP_FROM_ATTR`` class attribute.

        Args:
            h5grp (h5py.Group): The HDF5 group to create a dantro group for in
                the ``target`` group.
            target (BaseDataGroup): The group in which to create a new group
                that represents ``h5grp``
            name (str): the name of the new group
            GroupMap (dict): Map of names to BaseDataGroup-derived types;
                always needed, but may be empty
            map_attr (str): The HDF5 attribute to inspect in order to determine
                the name of the mapping
            **_: ignored
        """
        # Extract attributes manually
        attrs = {k: self._decode_attr_val(v) for k, v in h5grp.attrs.items()}

        # Determine the mapping type, falling back to the group default if no
        # mapping was specified
        _GroupCls = self._evaluate_type_mapping(
            map_attr, attrs=attrs, tmap=GroupMap, fallback=None
        )

        # Create and add the group, passing the attributes
        return target.new_group(path=name, Cls=_GroupCls, attrs=attrs)

    def _container_from_h5dataset(
        self,
        h5dset: "h5py.Dataset",
        target: BaseDataGroup,
        *,
        name: str,
        load_as_proxy: bool,
        proxy_kwargs: dict,
        DsetCls: type,
        map_attr: str,
        DsetMap: dict,
        plvl: int,
        pfstr: str,
        **_,
    ) -> BaseDataContainer:
        """Adds a new data container from a h5.Dataset

        The group types may be mapped to different dantro types; this is
        controlled by the extracted HDF5 attribute with the name specified in
        the ``_HDF5_MAP_FROM_ATTR`` class attribute.

        Args:
            h5dset (h5py.Dataset): The source dataset to load into ``target``
                as a dantro data container.
            target (BaseDataGroup): The target group where the ``h5dset`` will
                be represented in as a new dantro data container.
            name (str): the name of the new container
            load_as_proxy (bool): Whether to load as
                :py:class:`~dantro.proxy.hdf5.Hdf5DataProxy`
            proxy_kwargs (dict): Upon proxy initialization, unpacked into
                :py:meth:`dantro.proxy.hdf5.Hdf5DataProxy.__init__`
            DsetCls (BaseDataContainer): The type that is used to create
                the dataset-equivalents in ``target``. If mapping is enabled,
                this serves as the fallback type.
            map_attr (str): The HDF5 attribute to inspect in order to determine
                the name of the mapping
            DsetMap (dict): Map of names to BaseDataContainer-derived types;
                always needed, but may be empty
            plvl (int): the verbosity of the progress indicator
            pfstr (str): a format string for the progress indicator
        """
        # Progress information on loading this dataset
        if plvl >= 2:
            _info = pfstr.format(
                progress_info=self._progress_info_str,
                fpath=h5dset.file.filename,
                fname=os.path.basename(h5dset.file.filename),
                name=name,
                path=h5dset.name,
            )
            print_line(_info)

        # Extract attributes manually
        attrs = {k: self._decode_attr_val(v) for k, v in h5dset.attrs.items()}

        # Determine the class to use for this dataset.
        # If the target group supplies a default container type, specify None
        # as the fallback, delegating the choice to the `new_container` method.
        # Otherwise use the default specified here.
        _DsetCls = self._evaluate_type_mapping(
            map_attr,
            attrs=attrs,
            tmap=DsetMap,
            fallback=DsetCls if target._NEW_CONTAINER_CLS is None else None,
        )

        # Get the data, potentially as proxy.
        if load_as_proxy:
            data = Hdf5DataProxy(
                h5dset, **(proxy_kwargs if proxy_kwargs else {})
            )
        else:
            data = np.array(h5dset)

        # Now create and add the dataset
        return target.new_container(
            path=name, Cls=_DsetCls, data=data, attrs=attrs
        )

    # .........................................................................
    # Smaller helper methods

    def _decode_attr_val(self, attr_val) -> str:
        """Wrapper around decode_bytestrings"""
        # If feature not activated, return without doing anything
        if not self._HDF5_DECODE_ATTR_BYTESTRINGS:
            return attr_val

        return decode_bytestrings(attr_val)

    def _evaluate_type_mapping(
        self, key: str, *, attrs: dict, tmap: Dict[str, type], fallback: type
    ) -> type:
        """Given an attributes dict or group attributes, evaluates which type
        a target container should use.
        """

        def parse_map_attr(v) -> str:
            """Make sure the map attribute isn't a 1-sized array!"""
            if isinstance(v, np.ndarray):
                v = v.item()
            return str(v)

        # Easy cases first: no map or mapping attribute given
        if not tmap or not attrs.get(key):
            return fallback

        try:
            return tmap[parse_map_attr(attrs[key])]

        except KeyError:
            # Fall back to default
            log.warning(
                "Could not find a mapping from map attribute %s='%s' "
                "(originally %s) to a dantro container or group class. "
                "Available keys: %s. Using fallback type instead: %s .",
                key,
                parse_map_attr(attrs[key]),
                attrs[key],
                ", ".join([k for k in tmap]),
                fallback.__name__ if fallback is not None else "(none)",
            )
            return fallback



"""In this module, the :py:class:`~dantro.groups.graph.GraphGroup` is
implemented, which provides an interface between hierarchically stored data
and the creation of graph objects using the networkx package and the therein
implemented :py:class:`networkx.Graph` classes.

See :ref:`data_structures_graph_group` for more information.
"""

import logging
import warnings
from typing import List, Union

import numpy as np

from .._import_tools import LazyLoader
from ..containers import XrDataContainer
from ..tools import recursive_update
from . import BaseDataGroup, is_group
from .labelled import LabelledDataGroup

# Local constants
log = logging.getLogger(__name__)

# Lazy imports for packages that take a long time to import
nx = LazyLoader("networkx")
xr = LazyLoader("xarray")


# -----------------------------------------------------------------------------


@is_group
class GraphGroup(BaseDataGroup):
    """The GraphGroup class manages groups of graph data containers and
    provides the possibility to create networkx graph objects using the data
    inside this group.

    See :ref:`data_structures_graph_group` for more information.
    """

    # Define allowed member container types
    _ALLOWED_CONT_TYPES = (XrDataContainer, LabelledDataGroup)

    # Define, as class variables, in which containers or attributes to find the
    # info on the nodes and edges.
    _GG_node_container = "nodes"
    _GG_edge_container = "edges"
    _GG_attr_directed = "directed"
    _GG_attr_parallel = "parallel"
    _GG_attr_edge_container_is_transposed = "edge_container_is_transposed"
    _GG_attr_keep_dim = "keep_dim"

    # Define whether warning is raised upon bad alignment of property data
    _GG_WARN_UPON_BAD_ALIGN = True

    # .........................................................................

    def __init__(self, *args, **kwargs):
        """Initialize a GraphGroup.

        Args:
            *args: passed to :py:meth:`dantro.base.BaseDataGroup.__init__`
            **kwargs: passed to :py:meth:`dantro.base.BaseDataGroup.__init__`
        """
        super().__init__(*args, **kwargs)
        self._property_maps = dict()

    @property
    def property_maps(self) -> dict:
        """The property maps associated with this group, keyed by name."""
        return self._property_maps

    @property
    def node_container(self):
        """Returns the associated node container of this graph group"""
        try:
            return self[self._GG_node_container]

        except KeyError as err:
            raise KeyError(
                f"No container with name '{self._GG_node_container}' "
                f"available in {self.logstr}! Check if the class variable "
                "_GG_node_container is set to the correct value."
            ) from err

    @property
    def edge_container(self):
        """Returns the associated edge container of this graph group"""
        try:
            return self[self._GG_edge_container]

        except KeyError as err:
            raise KeyError(
                f"No container with name '{self._GG_edge_container}' "
                f"available in {self.logstr}! Check if the class variable "
                "_GG_edge_container is set to the correct value."
            ) from err

    @property
    def default_keep_dim(self):
        """The default dimensions not to be squeezed during data selection as
        specified in the respective group attribute.
        """
        if self._GG_attr_keep_dim in self.attrs:
            return tuple(self.attrs[self._GG_attr_keep_dim])
        return None

    # .........................................................................

    def _get_item_or_pmap(self, key: Union[str, List[str]]):
        """Returns the object accessible via ``key``. Apart from allowing to
        retrieve objects in this group, the method additionally allows to
        access data stored in property maps.

        Args:
            key (Union[str, List[str]]): The object to retrieve.
                If this is a path, will recurse down until at the end.

        Returns:
            The object at ``key``

        Raises:
            KeyError: If no such key can be found
        """
        # First, look for the key in the property maps
        if key in self._property_maps:
            return self._property_maps[key]

        # Else, invoke the parent method
        try:
            return super().__getitem__(key=key)

        except KeyError as err:
            _available_keys = [str(k) for k in self.keys()]
            _available_keys += [str(k) for k in self._property_maps.keys()]
            _available_keys = ", ".join(_available_keys)

            raise KeyError(
                f"No key, key sequence, or property '{key}' in {self.logstr}! "
                f"Available keys at top level: {_available_keys}"
            ) from err

    def _get_data_at(
        self,
        *,
        data: Union[XrDataContainer, LabelledDataGroup],
        sel: dict = None,
        isel: dict = None,
        at_time: int = None,
        at_time_idx: int = None,
        keep_dim=None,
    ) -> Union["xarray.DataArray", XrDataContainer]:
        """Returns a :py:class:`xarray.DataArray` containing the data
        specified via the selectors ``sel`` and ``isel``. Any dimension of
        size 1 is removed from the selected data.

        .. warning::

            Any invalid key in ``sel`` and ``isel`` is ignored silently.

        Args:
            data (Union[XrDataContainer, LabelledDataGroup]): Data to select
                from.
            sel (dict, optional): Dict of coordinate values keyed by
                dimensions, passed to ``data.sel``. Used to select data via
                index label. May be given together with ``isel`` if no key
                exists in both.
            isel (dict, optional): Dict of indexes keyed by dimensions,
                passed to ``data.isel``. Used to select data via index.
                May be given together with ``sel`` if no key exists in both.
            at_time (int, optional): Select along ``time`` dimension via index
                label. Translated to ``sel = dict(time=at_time)``, potentially
                overwriting an existing ``time`` entry.
            at_time_idx (int, optional): Select along ``time`` dimension via
                index. Translated to ``isel = dict(time=at_time_idx)``,
                potentially overwriting an existing ``time`` entry.
            keep_dim (optional): Iterable containing names of the dimensions
                that can not be squeezed.

        Returns:
            xarray.DataArray: The selected data

        Raises:
            ValueError: On keys that exist in both ``sel`` and ``isel``
        """
        # Update `sel` and `isel` with the explicit time specifications
        if at_time is not None:
            sel = recursive_update(sel if sel else {}, dict(time=at_time))

        if at_time_idx is not None:
            isel = recursive_update(
                isel if isel else {}, dict(time=at_time_idx)
            )

        # Check that no key is in both sel and isel
        if sel and isel:
            # Check that the intersection of the key-sets is empty
            if sel.keys() & isel.keys():
                _duplicate_keys = ", ".join(sel.keys() & isel.keys())
                raise ValueError(
                    "Received keys that appear in both `sel` and `isel` for "
                    f"the selection of {data}: {_duplicate_keys}"
                )

        log.debug(
            "Received the following selectors:\n  sel: %s\n  isel: %s",
            sel,
            isel,
        )
        log.debug("Now applying these to the following data ...\n  %s", data)

        # Get the dimensions available for selection
        dims = {d for d in data.dims}
        # If `data` is a group, extract the member level dimensions. It is
        # sufficient to look at a single group member, since dimensions along
        # which is selected have to be shared among all group members.
        if isinstance(data, BaseDataGroup):
            dims |= {d for d in data[next(iter(data))].dims}

        # Remove any entries for which no matching dimension exists and apply
        # sel _and_ isel (if given).
        if sel is not None:
            sel = {k: v for k, v in sel.items() if k in dims}
            data = data.sel(**sel)

        if isel is not None:
            isel = {k: v for k, v in isel.items() if k in dims}
            data = data.isel(**isel)

        log.debug(
            "Applied the following selectors:\n  sel: %s\nisel: %s", sel, isel
        )

        # If `data` is still a group, select everything to get a `xr.DataArray`
        # with the coordinates also from the member level.
        if isinstance(data, BaseDataGroup):
            data = data.sel()

        # `data` is now a `xr.DataArray`. Squeeze out any dimension of size 1
        # that is not in `keep_dim`.
        if keep_dim is None:
            keep_dim = (
                self.default_keep_dim
                if self.default_keep_dim is not None
                else []
            )

        for dim in data.dims:
            if data.sizes[dim] == 1 and dim not in keep_dim:
                data = data.squeeze(dim=dim)

        if not data.dims:
            # Oops, no dimension left. Add one default dimension.
            data = data.expand_dims("dim_0")

        return data

    def _prepare_edge_data(self, *, edges, max_tuple_size: int):
        """Prepares the edge data. Depending on the
        ``_GG_attr_edge_container_is_transposed`` class attribute, the edge
        data is transposed or not. If the attribute does not exist, the data is
        transposed only if the correct shape could unambiguously be deduced.

        Args:
            edges: The edge data stored in a 2-dimensional container
            max_tuple_size (int): The maximum allowed edge tuple size (4 for
                :py:class:`networkx.MultiGraph`, else 3). Used if the correct
                shape is tried to be deduced automatically.

        Returns:
            The edge data, possibly transposed

        Raises:
            TypeError: Edge data is not 2-dimensional
        """
        # Check that edga data is 2-dimensional (if not empty)
        if edges.values.size != 0 and edges.values.ndim != 2:
            _msg_squeezed_dims = ""

            if edges.values.ndim == 1:
                _msg_squeezed_dims = (
                    "\nThe edge dimension might have been squeezed during "
                    f"data selection. Use the `{self._GG_attr_keep_dim}` "
                    "group attribute or the `keep_dim` argument to specify "
                    "dimensions that are not to be squeezed."
                )

            raise TypeError(
                "Edge data must be 2-dimensional. Received edge data with "
                f"{edges.values.ndim} dimension(s). Also note that the "
                f"edge-tuples must be of equal size.{_msg_squeezed_dims}"
            )

        # If information on edge container shape is given as class attribute,
        # rely on that.
        if self._GG_attr_edge_container_is_transposed in self.attrs:
            if self.attrs[self._GG_attr_edge_container_is_transposed]:
                return edges.T
            return edges

        # Transpose if needed and if the correct shape is unambiguous,
        # i.e., if the size of one dimension lies not in [2, max_tuple_size].
        elif (
            not 2 <= edges.values.shape[-1] <= max_tuple_size
            and 2 <= edges.values.shape[0] <= max_tuple_size
        ):
            return edges.T

        return edges

    def _prepare_property_data(self, name: str, data):
        """Prepares external property data.

        Args:
            name (str): The properties' name
            data: The property data

        Returns:
            The data, potentially converted to a
                :py:class: `~dantro.containers.xr.XrDataContainer`

        Raises:
            TypeError: On invalid type of ``data``
        """
        if not any([isinstance(data, t) for t in self._ALLOWED_CONT_TYPES]):
            try:
                data = XrDataContainer(name=name, data=data)
            except:
                _allowed_types = ", ".join(
                    [str(t) for t in self._ALLOWED_CONT_TYPES]
                )
                raise TypeError(
                    f"Received invalid type for 'data' argument: {type(data)}."
                    f" Expected one of: {_allowed_types} ... or a type that "
                    "could be converted to an XrDataContainer."
                )

        return data

    def _check_alignment(self, *, ent, prop):
        """Checks the alignment of property data and entity (node or edge)
        data. If ``self._GG_WARN_UPON_BAD_ALIGN`` is True, warn on possible
        pitfalls.

        Args:
            ent: The entity (node or edge) data
            prop: The property data
        """
        if not self._GG_WARN_UPON_BAD_ALIGN:
            return

        # Check if any matching dimension (of size >1) exists
        if not any([d in prop.squeeze().dims for d in ent.squeeze().dims]):
            warnings.warn(
                "No matching dimensions found in property data "
                f"('{prop.name}') and entity data ('{ent.name}') for "
                "dimensions of size >1. No re-ordering was done during "
                "alinment.",
                UserWarning,
            )

        # Check for missing values in the property data after alignment
        if prop.isnull().any():
            warnings.warn(
                f"Found missing values in property data ('{prop.name}') after "
                f"alignment with entity data ('{ent.name}'). Make sure that "
                "the coordinate values are available in the property data.",
                UserWarning,
            )

    # .........................................................................

    def register_property_map(self, key: str, data):
        """Registers a new property map. It allows for the given data to be
        accessed internally by the specified key.

        Args:
            key (str): The key via which the registered data will be available
            data: The data to be mapped. If the given data is not an allowed
                container type, an attempt is made to construct an
                :py:class:`~dantro.containers.xr.XrDataContainer` with
                the data. Only if this operation fails, will property map
                registration fail.

        Raises:
            ValueError: On invalid key
        """
        data = self._prepare_property_data(key, data)

        # Ensure that the key does not exist already, which could lead to
        # unexpected shadowing of an existing group member.
        if key in self or key in self._property_maps:
            raise ValueError(
                f"The given key '{key}' conflicts with a valid member path in "
                f"{self.logstr}. Please choose a unique name."
            )

        # Everything ok, register the new map
        self._property_maps[key] = data

        log.remark(
            "Registered tag '%s' as property map of %s.", key, self.logstr
        )

    def create_graph(
        self,
        *,
        directed: bool = None,
        parallel_edges: bool = None,
        node_props: list = None,
        edge_props: list = None,
        sel: dict = None,
        isel: dict = None,
        at_time: int = None,
        at_time_idx: int = None,
        align: bool = False,
        keep_dim=None,
        **graph_kwargs,
    ) -> "networkx.Graph":
        """Create a networkx :py:class:`networkx.Graph` (or a more specialized
        graph type) object from the node and edge data associated with this
        graph group. Optionally, node and edge properties can be added from
        data stored or registered in the graph group.
        The coordinates for the selected or squeezed dimensions of the node,
        edge, and property data are stored as graph attributes in ``g.graph``.

        .. note::

            Any pre-selection specified by ``sel``, ``isel``, ``at_time``, or
            ``at_time_idx`` will be applied to the node data, edge data, as
            well as any given property data.

        .. warning::

            Any invalid key in ``sel`` and ``isel`` is ignored silently (see
            :py:meth:`~dantro.groups.graph.GraphGroup._get_data_at`).

        Args:
            directed (bool, optional): If true, the graph will be directed.
                If not given, the value given by the group attribute with name
                ``_GG_attr_directed`` is used instead.
            parallel_edges (bool, optional): If true, the graph will allow
                parallel edges. If not given, the value is tried to be read
                from the group attribute with name ``_GG_attr_parallel``.
            node_props (list, optional): List of names specifying the
                containers that contain the node property data.
            edge_props (list, optional): List of names specifying the
                containers that contain the edge property data.
            sel (dict, optional): Dict of coordinate values keyed by
                dimensions, passed to
                :py:meth:`~dantro.groups.graph.GraphGroup._get_data_at`.
                Used to select data via index label.
            isel (dict, optional): Dict of indexes keyed by dimensions,
                passed to
                :py:meth:`~dantro.groups.graph.GraphGroup._get_data_at`.
                Used to select data via index.
            at_time (int, optional): Select along ``time`` dimension via index
                label. Translated to ``sel = dict(time=at_time)``.
            at_time_idx (int, optional): Select along ``time`` dimension via
                index. Translated to ``isel = dict(time=at_time_idx)``.
            align (bool, optional): If True, the property data is aligned
                with the node/edge data using ``xarray.align``
                (default: False). The indexes of the ``<node/edge>_container``
                are used for each dimension. If the class variable
                ``_GG_WARN_UPON_BAD_ALIGN`` is True, warn upon missing values
                or if no re-ordering was done. Any dimension of size 1 is
                squeezed and thus alignment (via ``align=True``) will have no
                effect on such dimensions.
            keep_dim (optional): Iterable containing names of the dimensions
                that can not be squeezed. Passed on to
                :py:meth:`~dantro.groups.graph.GraphGroup._get_data_at`.
            **graph_kwargs: Passed to the constructor of the respective
                networkx graph object.

        Returns:
            The networkx graph object. Depending on the provided information,
            one of the following graph objects is created:
            :py:class:`networkx.Graph`, :py:class:`networkx.DiGraph`,
            :py:class:`networkx.MultiGraph`, :py:class:`networkx.MultiDiGraph`.
        """
        # Get the node and edge data stored in the graph group
        log.debug("Checking whether node and edge data is available...")

        node_cont = self._get_data_at(
            data=self.node_container,
            sel=sel,
            isel=isel,
            at_time=at_time,
            at_time_idx=at_time_idx,
            keep_dim=keep_dim,
        )
        edge_cont = self._get_data_at(
            data=self.edge_container,
            sel=sel,
            isel=isel,
            at_time=at_time,
            at_time_idx=at_time_idx,
            keep_dim=keep_dim,
        )

        # Get info on directed and parallel edges from attributes, if not
        # explicitly given
        if directed is None:
            directed = self.attrs[self._GG_attr_directed]

        if parallel_edges is None:
            parallel_edges = self.attrs[self._GG_attr_parallel]

        # Create a networkx graph corresponding to the graph properties
        log.debug("Creating a networkx graph object...")

        if not directed and not parallel_edges:
            g = nx.Graph(**graph_kwargs)

        elif directed and not parallel_edges:
            g = nx.DiGraph(**graph_kwargs)

        elif not directed and parallel_edges:
            g = nx.MultiGraph(**graph_kwargs)

        else:
            g = nx.MultiDiGraph(**graph_kwargs)

        # Prepare the edge data. If needed, the data is transposed
        edge_cont = self._prepare_edge_data(
            edges=edge_cont,
            max_tuple_size=(4 if isinstance(g, nx.MultiGraph) else 3),
        )

        # Drop missing values from the node-data and edge-data. After calling
        # `_prepare_edge_data`, the first edge-data dim should be the edge-dim.
        # This is also why the dropna kwarg can't be evaluated in _get_data_at.
        if np.isnan(node_cont).any():
            node_cont = node_cont.dropna(dim=node_cont.dims[0])
            log.debug(
                "Removed entries containg missing (NaN) values along dim '%s' "
                "from %s.",
                node_cont.dims[0],
                self.node_container.logstr,
            )

        if np.isnan(edge_cont).any():
            edge_cont = edge_cont.dropna(dim=edge_cont.dims[0])
            log.debug(
                "Removed entries containg missing (NaN) values along dim '%s' "
                "from %s.",
                edge_cont.dims[0],
                self.node_container.logstr,
            )

        # Add nodes and edges to the graph
        log.debug("Adding nodes to the graph...")
        g.add_nodes_from(node_cont.values)

        log.debug("Adding edges to the graph...")
        g.add_edges_from(edge_cont.values)

        # Store node and edge coordinates as graph attribute for dimensions
        # of size 1; they are discarded otherwise. Since the networkx graph
        # does not ensure the same edge-ordering, the edge-idx coordinates
        # should not be stored. For the sake of consistency, this is not done
        # for any of the dimensions of size > 1 (node-idx, edge-idx, edge).
        # If needed, the node-idx coordinates can be added as property.
        node_coords = {
            d: node_cont.coords[d].item()
            for d in node_cont.coords
            if node_cont.coords[d].size == 1
        }
        edge_coords = {
            d: edge_cont.coords[d].item()
            for d in edge_cont.coords
            if edge_cont.coords[d].size == 1
        }
        # The order of the dictionary update should not matter
        g.graph.update(node_coords, **edge_coords)

        # Add properties to nodes and edges
        if node_props:
            for prop_name in node_props:
                self.set_node_property(
                    g=g,
                    name=prop_name,
                    sel=sel,
                    isel=isel,
                    at_time=at_time,
                    at_time_idx=at_time_idx,
                    align=align,
                    keep_dim=keep_dim,
                )
        if edge_props:
            for prop_name in edge_props:
                self.set_edge_property(
                    g=g,
                    name=prop_name,
                    sel=sel,
                    isel=isel,
                    at_time=at_time,
                    at_time_idx=at_time_idx,
                    align=align,
                    keep_dim=keep_dim,
                )

        # Return the graph
        log.info(
            "Created graph (%d node%s, %d edge%s) from %s.",
            g.number_of_nodes(),
            "s" if g.number_of_nodes() != 1 else "",
            g.number_of_edges(),
            "s" if g.number_of_edges() != 1 else "",
            self.logstr,
        )

        return g

    def set_node_property(
        self,
        *,
        g: "networkx.Graph",
        name: str,
        data=None,
        align: bool = False,
        keep_dim=None,
        **selector,
    ):
        """Sets a property to every node in Graph ``g`` that is also in the
        ``node_container`` of the graph group. The coordinates for the selected
        or squeezed dimensions of the property data are stored as Graph
        attributes (in ``g.graph``).

        Args:
            g (networkx.Graph): The networkx graph object
            name (str): If ``data`` is ``None``, ``name`` must specify the
                container within the graph group that contains the property
                values, or be valid key in ``property_maps``. ``name`` is used
                as the name for the property in the graph object, potentially
                overwriting an existing property.
            data (None, optional): If given, load node properties directly
                from ``data``. If the given data is not an allowed container
                type, an attempt is made to construct an
                :py:class:`~dantro.containers.xr.XrDataContainer` with
                the data. Only if this operation fails, the node property
                setting will fail.
            align (bool, optional): If True, the property data is aligned
                with the node data using ``xarray.align``. The indexes of the
                ``node_container`` are used for each dimension. If the class
                variable ``_GG_WARN_UPON_BAD_ALIGN`` is True, warn upon
                missing values or if no re-ordering was done. Any dimension of
                size 1 is squeezed and thus alignment (via ``align=True``) will
                have no effect on such dimensions.
            keep_dim (optional): Iterable containing names of the dimensions
                that can not be squeezed. Passed on to
                :py:meth:`~dantro.groups.graph.GraphGroup._get_data_at`.
            **selector: Specifies the selection applied to both node data and
                property data. Passed on to
                :py:meth:`~dantro.groups.graph.GraphGroup._get_data_at`. Use
                the ``sel`` (``isel``) dict to select data via coordinate value
                (index).

        Raises:
            ValueError: Lenght mismatch of the selected property and node data
        """
        # Get the property data
        if data is None:
            prop_data = self._get_item_or_pmap(name)
        else:
            prop_data = self._prepare_property_data(name, data)

        prop_data = self._get_data_at(
            data=prop_data, keep_dim=keep_dim, **selector
        )

        # Get the node data
        node_cont = self._get_data_at(
            data=self.node_container, keep_dim=keep_dim, **selector
        )

        if np.isnan(node_cont).any():
            node_cont = node_cont.dropna(dim=node_cont.dims[0])
            log.debug(
                "Removed entries containg missing (NaN) values along dim '%s' "
                "from %s",
                node_cont.dims[0],
                self.node_container.logstr,
            )

        # Optionally, align the property data with the node data.
        if align:
            node_cont, prop_data = xr.align(node_cont, prop_data, join="left")
            self._check_alignment(ent=node_cont, prop=prop_data)

        # Check if the data can be added as node property
        if len(prop_data) != len(node_cont):
            # Prepare error message
            _msg_squeezed_dims = ""
            _msg_match_with_node_number = ""

            if len(g.nodes) == 1:
                _msg_squeezed_dims = (
                    "\n"
                    "The node dimension in the property data might have been "
                    "squeezed. Use the `keep_dim` argument to specify "
                    "dimensions not to be squeezed."
                )

            if len(g.nodes) == len(prop_data):
                _msg_match_with_node_number = (
                    "\n"
                    f"The size of '{name}' matches the current number of "
                    "nodes, which changed since the graph creation. However, "
                    "properties can only be added to those nodes that are in "
                    f"{self.node_container.logstr}. Alternatively, add them "
                    "manually using nx.set_node_attributes."
                )

            raise ValueError(
                f"Length mismatch: Failed to add '{name}' data as a node "
                f"property. Received {len(prop_data)} property values for "
                f"{len(node_cont)} nodes in {self.node_container.logstr}! "
                f"{_msg_match_with_node_number}{_msg_squeezed_dims}"
            )

        n_data = len(node_cont)
        n_graph = len(g.nodes)

        if n_data != n_graph:
            warnings.warn(
                "The number of nodes "
                f"{'decreased' if n_graph < n_data else 'increased'} since "
                f"graph creation (length of {self.node_container.logstr}: "
                f"{n_data}, nodes in graph: {n_graph}). Some property values "
                "will not be added to the graph! Reasons might be previous "
                "modifications of the graph structure, duplicate entries in "
                f"'{self.node_container.logstr}', or nodes in "
                f"'{self.edge_container.logstr}' that are not in "
                f"'{self.node_container.logstr}'.",
                UserWarning,
            )

        # Create dict of properties keyed by node
        props = {n: val for n, val in zip(node_cont.values, prop_data.values)}

        # Add property to graph
        nx.set_node_attributes(g, values=props, name=name)

        # Store the coords of the selected property data for dimensions of size
        # 1 as graph attribute (potentially overwriting existing attributes).
        prop_coords = {
            d: prop_data.coords[d].item()
            for d in prop_data.coords
            if prop_data.coords[d].size == 1
        }
        g.graph.update(prop_coords)

        log.remark("Node property '%s' added from %s.", name, self.logstr)

    def set_edge_property(
        self,
        *,
        g: "networkx.Graph",
        name: str,
        data=None,
        align: bool = False,
        keep_dim=None,
        **selector,
    ):
        """Sets a property to every edge in Graph ``g`` that is also in the
        ``edge_container`` of the graph group. The coordinates for the selected
        or squeezed dimensions of the property data are stored as Graph
        attributes (in ``g.graph``).

        Args:
            g (networkx.Graph): The networkx graph object
            name (str): If ``data`` is ``None``, ``name`` must specify the
                container within the graph group that contains the property
                values, or be valid key in ``property_maps``. ``name`` is used
                as the name for the property in the graph object, potentially
                overwriting an existing property.
            data (None, optional): If given, load edge properties directly
                from ``data``. If the given data is not an allowed container
                type, an attempt is made to construct an
                :py:class:`~dantro.containers.xr.XrDataContainer` with
                the data. Only if this operation fails, the edge property
                setting will fail.
            align (bool, optional): If True, the property data is aligned
                with the edge data using ``xarray.align``. The indexes of the
                ``edge_container`` are used for each dimension. If the class
                variable ``_GG_WARN_UPON_BAD_ALIGN`` is True, warn upon
                missing values or if no re-ordering was done. Any dimension of
                size 1 is squeezed and thus alignment (via ``align=True``) will
                have no effect on such dimensions.
            keep_dim (optional): Iterable containing names of the dimensions
                that can not be squeezed. Passed on to
                :py:meth:`~dantro.groups.graph.GraphGroup._get_data_at`.
            **selector: Specifies the selection applied to both edge data and
                property data. Passed on to
                :py:meth:`~dantro.groups.graph.GraphGroup._get_data_at`. Use
                the ``sel`` (``isel``) dict to select data via coordinate value
                (index).

        Raises:
            ValueError: Lenght mismatch of the selected property and edge data
        """

        def new_edge_key(e, edges, directed: bool):
            """Returns an unused edge key.

            Finds the last edge in `edges` that represents the same edge as
            ``e`` and return its key incremented by 1.
            If `e` is not found, it is the first being added. Thus, return 0.
            """
            e = tuple(e)

            if directed:
                match = lambda edge: e == edge[:2]
            else:
                match = lambda edge: e == edge[:2] or e == edge[:2][::-1]

            for edge in edges[::-1]:
                if match(edge):
                    return edge[2] + 1
            # First key to add, may return 0
            return 0

        # Get the property data
        if data is None:
            prop_data = self._get_item_or_pmap(name)
        else:
            prop_data = self._prepare_property_data(name, data)

        prop_data = self._get_data_at(
            data=prop_data, keep_dim=keep_dim, **selector
        )

        # Get and prepare the edge data
        edge_cont = self._get_data_at(
            data=self.edge_container, keep_dim=keep_dim, **selector
        )
        edge_cont = self._prepare_edge_data(
            edges=edge_cont,
            max_tuple_size=(4 if isinstance(g, nx.MultiGraph) else 3),
        )

        if np.isnan(edge_cont).any():
            edge_cont = edge_cont.dropna(dim=edge_cont.dims[0])
            log.debug(
                "Removed entries containg missing (NaN) values along dim '%s' "
                "from %s",
                edge_cont.dims[0],
                self.node_container.logstr,
            )

        # Optionally, align the property data with the node data.
        if align:
            edge_cont, prop_data = xr.align(edge_cont, prop_data, join="left")
            self._check_alignment(ent=edge_cont, prop=prop_data)

        # Check if the data can be added as edge property
        if len(prop_data) != len(edge_cont):
            # Prepare error message
            _msg_duplicate_edges = ""
            _msg_match_with_edge_number = ""

            if type(g) is nx.DiGraph or type(g) is nx.Graph:
                _msg_duplicate_edges = (
                    "\n"
                    f"Multiple entries in {self.edge_container.logstr} might "
                    "be interpreted as a single one, due to the chosen graph "
                    f"type: '{type(g)}'"
                )

            if len(g.edges) == len(prop_data):
                _msg_match_with_edge_number = (
                    "\n"
                    f"The size of '{name}' matches the current number of "
                    "edges, which changed since the graph creation. However, "
                    "properties can only be added to those edges that are in "
                    f"{self.edge_container.logstr}. Alternatively, add them "
                    "manually using nx.set_edge_attributes."
                )

            raise ValueError(
                f"Length mismatch! Failed to add '{name}' data as an edge "
                f"property. Received {len(prop_data)} property values for "
                f"{len(edge_cont)} edges in {self.edge_container.logstr}! "
                f"{_msg_duplicate_edges}{_msg_match_with_edge_number}"
            )

        n_data = len(edge_cont)
        n_graph = len(g.edges)

        if n_data != n_graph:
            warnings.warn(
                "The number of edges "
                f"{'decreased' if n_graph < n_data else 'increased'} since "
                f"graph creation (length of {self.edge_container.logstr}: "
                f"{n_data}, edges in graph: {n_graph}). Some property values "
                "will not be added to the graph! Reasons might be previous "
                "modifications of the graph structure or duplicate entries in "
                f"{self.edge_container.logstr}.",
                UserWarning,
            )

        # Now, prepare creation of dict of properties keyed by edge. If the
        # graph type does not allow for parallel edges, the keys (edges) must
        # be 2-tuples. Else, they must be 3-tuples, additionally containing an
        # edge key.
        # Thus, in the case of parallel edges: If edge data contains 2-tuples,
        # create the same edge keys as it was done during graph creation.
        if isinstance(g, nx.MultiGraph):
            edge_length = edge_cont.values.shape[-1]

            if edge_length == 2:
                edges = []
                for e in edge_cont:
                    edges.append(
                        (
                            int(e[0]),
                            int(e[1]),
                            new_edge_key(e, edges, g.is_directed()),
                        )
                    )

            else:
                edges = [tuple(e) for e in edge_cont.values]

        else:
            edges = [tuple(e) for e in edge_cont.values]

        # Create dict of property values keyed by edge
        props = {n: val for n, val in zip(edges, prop_data.values)}

        # Add property to graph
        nx.set_edge_attributes(g, values=props, name=name)

        # Store the coords of the selected property data for dimensions of size
        # 1 as graph attribute (potentially overwriting existing attributes).
        prop_coords = {
            d: prop_data.coords[d].item()
            for d in prop_data.coords
            if prop_data.coords[d].size == 1
        }
        g.graph.update(prop_coords)

        log.remark("Edge property '%s' added from %s.", name, self.logstr)



"""Implements the data operations database of dantro, which is used in
the :ref:`data transformation framework <dag_framework>` to apply
transformations on data using :py:class:`~dantro.dag.TransformationDAG`.

isort:skip_file
"""

from .db import _OPERATIONS
from .db_tools import (
    available_operations,
    get_operation,
    register_operation,
    is_operation,
)
from .apply import apply_operation
from .hooks import DAG_PARSER_OPERATION_HOOKS



"""Implements data operations that work with expressions, e.g. lambda function
definitions or symbolic math"""

import logging
import math
import re
from typing import Callable, Tuple, Union

import numpy as np

from .._import_tools import LazyLoader

log = logging.getLogger(__name__)

xr = LazyLoader("xarray")
nx = LazyLoader("networkx")
pd = LazyLoader("pandas")
scipy = LazyLoader("scipy")

# -----------------------------------------------------------------------------


def expression(
    expr: str,
    *,
    symbols: dict = None,
    evaluate: bool = True,
    transformations: Tuple[Callable] = None,
    astype: Union[type, str] = float,
):
    """Parses and evaluates a symbolic math expression using SymPy.

    For parsing, uses sympy's :py:func:`sympy.parsing.sympy_parser.parse_expr`.
    The ``symbols`` are provided as ``local_dict``; the ``global_dict`` is not
    explicitly set and subsequently uses the sympy default value, containing
    all basic sympy symbols and notations.

    .. note::

        The expression given here is not Python code, but symbolic math.
        You cannot call arbitrary functions, but only those that are imported
        by ``from sympy import *``.

    .. hint::

        When using this expression as part of the :ref:`dag_framework`, it is
        attached to a so-called :ref:`syntax hook <dag_op_hooks_integration>`
        that makes it easier to specify the ``symbols`` parameter.
        See :ref:`here <dag_op_hook_expression>` for more information.

    .. warning::

        While the expression is symbolic math, be aware that smypy by default
        interprets the ``^`` operator as XOR, not an exponentiation!
        For exponentiation, use the ``**`` operator or adjust the
        ``transformations`` argument as specified in the sympy documentation.

    .. warning::

        The return object of this operation will *only* contain symbolic sympy
        objects if ``astype is None``. Otherwise, the type cast will evaluate
        all symbolic objects to the numerical equivalent specified by the given
        ``astype``.

    Args:
        expr (str): The expression to evaluate
        symbols (dict, optional): The symbols to use
        evaluate (bool, optional): Controls whether sympy evaluates ``expr``.
            This *may* lead to a fully evaluated result, but does not guarantee
            that no sympy objects are contained in the result. For ensuring
            a fully numerical result, see the ``astype`` argument.
        transformations (Tuple[Callable], optional): The ``transformations``
            argument for sympy's
            :py:func:`sympy.parsing.sympy_parser.parse_expr`. By default, the
            sympy standard transformations are performed.
        astype (Union[type, str], optional): If given, performs a cast to this
            data type, fully evaluating all symbolic expressions.
            Default: Python :py:class:`float`.

    Raises:
        TypeError: Upon failing ``astype`` cast, e.g. due to free symbols
            remaining in the evaluated expression.
        ValueError: When parsing of ``expr`` failed.

    Returns:
        The result of the evaluated expression.
    """
    from sympy.parsing.sympy_parser import parse_expr as _parse_expr
    from sympy.parsing.sympy_parser import standard_transformations as _std_trf

    log.remark("Evaluating symbolic expression:  %s", expr)

    symbols = symbols if symbols else {}
    parse_kwargs = dict(
        evaluate=evaluate,
        transformations=(
            transformations if transformations is not None else _std_trf
        ),
    )

    # Now, parse the expression
    try:
        res = _parse_expr(expr, local_dict=symbols, **parse_kwargs)

    except Exception as exc:
        raise ValueError(
            f"Failed parsing expression '{expr}'! Got a "
            f"{exc.__class__.__name__}: {exc}. Check that the expression can "
            "be evaluated with the available symbols "
            f"({', '.join(symbols) if symbols else 'none specified'}) "
            "and inspect the chained exceptions for more information. Parse "
            f"arguments were: {parse_kwargs}"
        ) from exc

    # Finished here if no type cast is desired
    if astype is None:
        return res

    # If full evaluation is desired, do so via a numpy type cast. This works on
    # all sympy objects, but _importantly_ also works on numpy arrays
    # containing sympy objects.
    dtype = np.dtype(astype)
    log.debug("Applying %s ...", dtype)

    try:
        return dtype.type(res)

    except Exception as exc:
        raise TypeError(
            f"Failed casting the result of expression '{expr}' from "
            f"{type(res)} to {dtype}! This can also be due to free symbols "
            "remaining in the evaluated expression. Either specify the free "
            f"symbols (got: {', '.join(symbols) if symbols else 'none'}) or "
            "deactivate casting by specifying None as ``dtype`` argument. "
            f"The expression evaluated to:\n\n    {res}\n"
        ) from exc


def generate_lambda(expr: str) -> Callable:
    """Generates a lambda from a string. This is useful when working with
    callables in other operations.

    The ``expr`` argument needs to be a valid Python
    `lambda expression <https://docs.python.org/3/reference/expressions.html#lambda>`_.

    Inside the lambda body, the following names are available for use:

        * A large part of the ``builtins`` module
        * Every name from the Python ``math`` module, e.g. ``sin``, ``cos``, …
        * These modules (and their long form): ``np``, ``xr``, ``scipy``

    Internally, this uses ``eval`` but imposes the following restrictions:

        * The following strings may *not* appear in ``expr``: ``;``, ``__``.
        * There can be no nested ``lambda``, i.e. the only allowed lambda
          string is that in the beginning of ``expr``.
        * The dangerous parts from the ``builtins`` module are *not* available.

    Args:
        expr (str): The expression string to evaluate into a lambda.

    Returns:
        Callable: The generated Callable.

    Raises:
        SyntaxError: Upon failed evaluation of the given expression, invalid
            expression pattern, or disallowed strings in the lambda body.
    """
    ALLOWED_BUILTINS = (
        "abs",
        "all",
        "any",
        "callable",
        "chr",
        "divmod",
        "format",
        "hash",
        "hex",
        "id",
        "isinstance",
        "issubclass",
        "iter",
        "len",
        "max",
        "min",
        "next",
        "oct",
        "ord",
        "print",
        "repr",
        "round",
        "sorted",
        "sum",
        "None",
        "Ellipsis",
        "False",
        "True",
        "bool",
        "bytes",
        "bytearray",
        "complex",
        "dict",
        "enumerate",
        "filter",
        "float",
        "frozenset",
        "int",
        "list",
        "map",
        "range",
        "reversed",
        "set",
        "slice",
        "str",
        "tuple",
        "type",
        "zip",
        "open",
    )
    DISALLOWED_STRINGS = ("lambda", ";", "__")
    LAMBDA_PATTERN = r"^\s*lambda\s[\w\s\,\*]+\:(.+)$"
    #                   arguments definitions  : lambda body (capture group)
    # See also:  https://regex101.com/r/OmI8NY/2/

    # Check if the given expression matches this pattern
    pattern = re.compile(LAMBDA_PATTERN)
    match = pattern.match(expr)
    if match is None:
        raise SyntaxError(
            f"The given expression '{expr}' was not a valid lambda expression!"
        )

    # Check if the lambda body contains disallowed strings
    lambda_body = match[1]
    if any([bad_str in lambda_body for bad_str in DISALLOWED_STRINGS]):
        raise SyntaxError(
            "Encountered one or more disallowed strings in the "
            f"body ('{lambda_body}') of the given lambda "
            f"expression ('{expr}'). Make sure none of the "
            "following strings appears there: "
            f"{DISALLOWED_STRINGS}"
        )

    # Ok, sanitized enough now. Prepare the globals dict, restricting access to
    # a subset of builtins and only allowing commonly used math functionality
    _g = dict(
        __builtins__={
            n: f for n, f in __builtins__.items() if n in ALLOWED_BUILTINS
        },
        math=math,
        scipy=scipy,
        pandas=pd,
        numpy=np,
        xarray=xr,
        networkx=nx,
        np=np,
        xr=xr,
        nx=nx,
        pd=pd,
        **{n: f for n, f in math.__dict__.items() if not n.startswith("_")},
    )

    # Try evaluation now (with empty locals)
    try:
        f = eval(expr, _g, {})

    except Exception as exc:
        raise SyntaxError(
            "Failed generating lambda object from expression "
            f"'{expr}'! Got a {exc.__class__.__name__}: {exc}"
        ) from exc

    return f



"""Implements operations that need to be importable from other modules or that
are so basic that they apply to a wide range of applications."""

import operator
from typing import Callable

# -----------------------------------------------------------------------------

# fmt: off
BOOLEAN_OPERATORS = {
    "==": operator.eq,  "eq": operator.eq,
    "<":  operator.lt,  "lt": operator.lt,
    "<=": operator.le,  "le": operator.le,
    ">":  operator.gt,  "gt": operator.gt,
    ">=": operator.ge,  "ge": operator.ge,
    "!=": operator.ne,  "ne": operator.ne,
    "^":  operator.xor, "xor": operator.xor,
    #
    # Expecting an iterable as second argument
    "contains":         operator.contains,
    "in":               (lambda x, y: x in y),
    "not in":           (lambda x, y: x not in y),
    #
    # Performing bitwise boolean operations to support numpy logic
    "in interval":      (lambda x, y: x >= y[0] & x <= y[1]),
    "not in interval":  (lambda x, y: x < y[0] | x > y[1]),
} # End of boolean operator definitions
"""Boolean binary operators"""
# fmt: on


# -----------------------------------------------------------------------------


def _make_passthrough(func: Callable) -> Callable:
    """Wraps a callable such that it returns its first positional argument.

    This is meant for functions that operate on an object (conventionally the
    first argument) and do not have a return value.
    By constructing a callable using this function, it can be made compatible
    with the data transformation framework.

    .. code-block:: python

        f = setattr                 # f has no return value
        g = _make_passtrough(f)     # g will return the first argument
    """

    def wrapped(d, *args, **kwargs):
        func(d, *args, **kwargs)
        return d

    return wrapped



"""Implements :py:class:`~dantro.groups.labelled.LabelledDataGroup`
specializations for time series data.
"""

from . import is_group
from .labelled import LabelledDataGroup
from .ordered import IndexedDataGroup

# -----------------------------------------------------------------------------


@is_group
class TimeSeriesGroup(LabelledDataGroup, IndexedDataGroup):
    """A time-series group assumes that each stored member refers to one
    point in time, where the name is to be interpreted as the time coordinate.

    For more information on selection methods, see:

        * :py:meth:`~dantro.groups.labelled.LabelledDataGroup.sel`
        * :py:meth:`~dantro.groups.labelled.LabelledDataGroup.isel`
    """

    LDG_DIMS = ("time",)
    """Expected dimension names. There is only one dimension in a
    TimeSeriesGroup: ``time``"""

    LDG_EXTRACT_COORDS_FROM = "name"
    """Where to extract time coordinates from. Here, the container name is
    expected to be the time coordinate."""


@is_group
class HeterogeneousTimeSeriesGroup(TimeSeriesGroup):
    """This extends the :py:class:`~dantro.groups.time_series.TimeSeriesGroup`
    by configuring it such that it retrieves its coordinates not from the name
    of the members contained in it but from _their_ data.

    It still manages only the ``time`` dimension, which is now overlapping with
    the ``time`` dimension in the members of this group. However, the
    py:class:`~dantro.groups.labelled.LabelledDataGroup` can handle this
    overlap and provides a uniform selection interface that allows combining
    this heterogeneously stored data.

    This becomes especially useful in cases where the members of this group
    store data with the following properties:

        * Potentially different coordiantes than the coordinates of other
          members of the group.
        * Containing time information for more than a single time coordinate
        * No guarantee for overlaps between ``time`` dimension or any other
          dimension.

    As such it is suitable to work with data that represents ensembles that
    frequently change not only their size but also their identifying labels.
    Additionally, it supports them not being stored in regular intervals but
    only upon a change in coordinates.
    """

    LDG_EXTRACT_COORDS_FROM = "data"
    """Where to extract coordinates from. Here, coordinates are extracted from
    the data directly, inspecting only the group-level dimensions (``time``).
    """



"""The plotting module consolidates the dantro plotting framework

isort:skip_file
"""

from .creators import *
from ..exceptions import SkipPlot
from .plot_helper import PlotHelper
from .utils.plot_func import is_plot_func
from .utils.color_mngr import ColorManager, parse_cmap_and_norm_kwargs



"""Defines a loader mixin to load data via :py:mod:`pandas`"""

from ..containers import PassthroughContainer
from ._registry import add_loader

# -----------------------------------------------------------------------------


class PandasLoaderMixin:
    """Supplies functionality to load data via :py:mod:`pandas`."""

    @add_loader(TargetCls=PassthroughContainer)
    def _load_pandas_csv(
        filepath: str,
        *,
        TargetCls: type,
        **load_kwargs,
    ) -> PassthroughContainer:
        """Loads CSV data using :py:func:`pandas.read_csv`, returning a
        :py:class:`~dantro.containers.general.PassthroughContainer`
        that contains a :py:class:`pandas.DataFrame`.

        .. note::

            As there is no proper equivalent of a :py:class:`pandas.DataFrame`
            in dantro (yet), and unpacking the dataframe into a dantro group
            would reduce functionality, a passthrough-container is used here.
            It behaves mostly like the object it wraps.

            However, in some cases, you may have to retrieve the underlying
            data using the ``.data`` property.

        Args:
            filepath (str): Where the CSV data file is located
            TargetCls (type): The class constructor
            **load_kwargs: Passed on to :py:func:`pandas.read_csv`

        Returns:
            PassthroughContainer: Payload being the loaded CSV data in form of
                a :py:class:`pandas.DataFrame`.
        """
        import pandas as pd

        df = pd.read_csv(filepath, **load_kwargs)

        return TargetCls(data=df, attrs=dict(filepath=filepath))

    @add_loader(TargetCls=PassthroughContainer)
    def _load_pandas_generic(
        filepath: str,
        *,
        TargetCls: type,
        reader: str,
        **load_kwargs,
    ) -> PassthroughContainer:
        """Loads data from a file using one of :py:mod:`pandas` ``read_*``
        functions, returning a :py:class:`pandas.DataFrame` wrapped into a
        :py:class:`~dantro.containers.general.PassthroughContainer`.

        The ``reader`` argument needs to match a reader function from
        `pandas IO <https://pandas.pydata.org/pandas-docs/stable/user_guide/io.html>`_.

        .. note::

            As there is no proper equivalent of a :py:class:`pandas.DataFrame`
            in dantro (yet), and unpacking the dataframe into a dantro group
            would reduce functionality, a passthrough-container is used here.
            It behaves mostly like the object it wraps.

            However, in some cases, you may have to retrieve the underlying
            data using the ``.data`` property.

        .. note::

            Some of pandas' reader functions require additional packages to
            have been installed.

        .. warning::

            While this in principle allows access to reader functions that are
            *not* file-based, calling those will most probably fail because the
            functions do not expect a file path as their first argument.

        Args:
            filepath (str): Where the data file is located
            TargetCls (type): The class constructor
            reader (str): The name of the reader function from pandas IO to use
            **load_kwargs: Passed on to the reader function

        Returns:
            PassthroughContainer: Payload being the loaded data in form of
                a :py:class:`pandas.DataFrame`.
        """
        import pandas as pd

        try:
            read_func = getattr(pd, f"read_{reader}")

        except AttributeError as err:
            NOT_FILE_READERS = (
                "clipboard",
                "gbq",
                "sql",
                "sql_query",
                "sql_table",
            )
            _avail = ", ".join(
                s[5:]
                for s in dir(pd)
                if s.startswith("read_") and s[5:] not in NOT_FILE_READERS
            )
            raise ValueError(
                f"Invalid pandas reader name '{reader}'!\n"
                f"Available readers:  {_avail}"
            ) from err

        df = read_func(filepath, **load_kwargs)

        return TargetCls(data=df, attrs=dict(filepath=filepath))



"""A module containing tools for generating plot configurations"""

import copy
import logging
from collections import OrderedDict
from difflib import get_close_matches as _get_close_matches
from itertools import chain as _chain
from typing import Dict, List, Sequence, Tuple, Union

from paramspace import ParamSpace

from ..exceptions import PlotConfigError
from ..tools import make_columns, recursive_update

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


def _check_visited(
    visited: Sequence[Tuple[str, str]], *, next_visit: Tuple[str, str]
) -> Sequence[Tuple[str, str]]:
    """Performs cycle detection on the sequence of visited entries and raises
    an error if there will be a cycle. Otherwise, returns the new visiting
    sequence by appending the ``next_visit`` to the given sequence of
    ``visited`` entries.
    """
    if next_visit in visited:
        _loop = " <- ".join(
            [f"{b}::{p}" for b, p in _chain(visited, (next_visit,))]
        )
        raise PlotConfigError(
            f"While resolving the plot configuration for plot "
            f"'{visited[0][1]}', detected a circular dependency:  {_loop}  "
            "(with arrows denoting dependency and, plot configurations "
            "labelled in the form <base config label>::<plot name>). "
            "Check the `based_on` entries of the involved plot configurations "
            "and make sure that the determined configurations come from the "
            "*intended* base configuration."
        )
    return tuple(visited) + (next_visit,)


def _find_in_pool(
    name: str,
    *,
    base_pools: OrderedDict,
    skip: Sequence[Tuple[str, str]] = (),
) -> Tuple[str, dict, Tuple[str, dict]]:
    """Looks up a plot configuration in the given pool and returns the name
    of the pool, the found configuration, and the subset of pools that were not
    yet looked up.

    With ``skip``, certain entries can be skipped, e.g. the entry from which
    the current ``based_on`` is resolved from.
    """
    for i, (pool_name, pool_cfgs) in enumerate(reversed(base_pools.items())):
        if (pool_name, name) in skip or pool_cfgs is None:
            log.debug("Skipping '%s::%s'...", pool_name, name)
            continue

        try:
            pcfg = pool_cfgs[name]
            log.debug("Found '%s' in pool '%s'.", name, pool_name)
            break
        except KeyError:
            pass

    else:
        # Failed to find one. Generate a useful error message
        all_names = set(
            _chain(
                *[pc.keys() for _, pc in base_pools.items() if pc is not None]
            )
        )
        matches = _get_close_matches(name, all_names, n=5)
        _dym = ""
        if matches:
            _dym = f"Did you mean: {', '.join(matches)} ?\n"

        _pools = "\n".join(
            [
                f"--- Pool '{name}'\n{make_columns(pc.keys())}"
                for name, pc in reversed(base_pools.items())
                if pc is not None
            ]
        )
        raise PlotConfigError(
            f"Did not find a base plot configuration named '{name}' in the "
            f"pool of available base configurations! {_dym}Check that an "
            "entry with that name is part of at least one of the specified "
            f"pools:\n{_pools}"
        )

    # Found the desired configuration by searching the last i entries.
    # Reduce the base_pools to the subset that was not yet searched and the
    # currently used one. Then have all information ready to return ...
    _base_pools = (
        OrderedDict(list(base_pools.items())[:-i]) if i > 0 else base_pools
    )
    return pool_name, pcfg, _base_pools


def _resolve_based_on(
    pcfg: Union[dict, ParamSpace],
    *,
    base_pools: OrderedDict,
    _visited: Sequence[Tuple[str, str]],
) -> Union[dict, ParamSpace]:
    """Assembles a single plot's configuration by recursively resolving its
    ``based_on`` entry from a pool of available base plot configurations.

    This function *always* works on a deep copy of ``pcfg`` and will remove any
    potentially existing ``based_on`` entry on the root level of ``pcfg``.

    Furthermore, it accepts ``ParamSpace`` objects for plot configuration
    entries, recursively updating their dict representation and again creating
    a ``ParamSpace`` object from them afterwards.
    """
    # Need a few helper functions to handle ParamSpace objects
    # ... this COULD be avoided if ParamSpace would behave more dict-like,
    #     which it currently does not, so we have no choice but to unpack and
    #     repack it all the time. Should not be a big issue though, compared to
    #     all the plotting operations.
    _from_pspace = isinstance(pcfg, ParamSpace)
    _generate_return_value = lambda d: d if not _from_pspace else ParamSpace(d)
    _unpack_pspace = lambda d: d if not isinstance(d, ParamSpace) else d._dict
    # FIXME Should not have to use private API!

    # Prepare the given configuration
    pcfg = _unpack_pspace(copy.deepcopy(pcfg))

    based_on = pcfg.pop("based_on", None)
    if not based_on:
        return _generate_return_value(pcfg)

    elif isinstance(based_on, str):
        based_on = (based_on,)

    # Aggregate the based_on entries
    _pcfg = dict()
    for _based_on in based_on:
        log.debug("Resolving based_on: '%s' ...", _based_on)

        pool_name, base_cfg, sub_base_pools = _find_in_pool(
            _based_on, base_pools=base_pools, skip=(_visited[-1],)
        )

        # Might need to recursively resolve entries on that one ...
        # NOTE This also ensures that `base_cfg` is a deep copy, removing any
        #      possible mutability side effects from the recursive_update below
        base_cfg = _resolve_based_on(
            base_cfg,
            base_pools=sub_base_pools,
            _visited=_check_visited(
                _visited, next_visit=(pool_name, _based_on)
            ),
        )
        base_cfg = _unpack_pspace(base_cfg)

        # ... now apply to existing configuration
        _pcfg = recursive_update(_pcfg, base_cfg)

    # Finally, apply the given top level configuration
    pcfg = recursive_update(_pcfg, pcfg)
    return _generate_return_value(pcfg)


def resolve_based_on(
    plots_cfg: Dict[str, Union[dict, ParamSpace]],
    *,
    label: str,
    base_pools: Union[OrderedDict, Sequence[Tuple[str, dict]]],
) -> Dict[str, dict]:
    """Resolves the ``based_on`` entries of *all* plot configurations in the
    given plot configurations dictionary ``plots_cfg``.

    The procedure is as follows:

        - Iterate over root-level entries in the ``plots_cfg`` dict
        - For each entry, check if a ``based_on`` entry is present and needs
          to be resolved.
        - If so, recursively resolve the configuration entry, starting from the
          first entry in the ``based_on`` sequence and recursively updating it
          with content of the following elements. The final recursive update
          is that of the plot configuration given in ``plots_cfg``.

    Lookups happen from a pool of plot configurations: the ``base_pools``,
    combined with the given ``plots_cfg`` itself. The ``based_on`` entries are
    looked up by name using the following rules:

        - Other plot configurations within ``plots_cfg`` have highest
          precedence.
        - If no name is found there, lookups happen from within ``base_pools``,
          iterating over it *in reverse*, meaning that entries later in the
          ordered dict take precedence over those earlier.
        - If entries in a base pool are again using ``based_on``, these will be
          looked up using the same rules, but with the pool restricted to
          entries *with lower precedence* than that pool.
        - Lookups within the same pool will exclude the name of the currently
          updated plot configuration.
          Example: ``some_plot: {based_on: some_plot}`` will look for
          ``some_plot`` in some lower-precedence pool.

    The resolution of plot configurations works on deep copies of the given
    ``plots_cfg`` and all the ``based_on`` entries to avoid mutability issues
    between parts of these highly nested dictionaries.

    For integrated use of this functionality, see :ref:`plot_cfg_inheritance`.

    Args:
        plots_cfg (dict): A dict with multiple plot configurations to resolve
            the ``based_on`` entries of. Root-level keys are assumed to
            correspond to individual plot configurations.
            If this argument evaluates to False, will silently assume an empty
            plots configuration.
        label (str): The label to use for the given plots configuration when
            adding it to the base configuration pool.
        base_pools (Union[OrderedDict, Sequence[Tuple[str, dict]]]): The base
            configuration pools to look up the ``based_on`` entries in. This
            needs to be an OrderedDict or a type that can be converted into
            one. Keys will be used as labels for the individual pools.
            The order of this pool is relevant, see above.

    Raises:
        PlotConfigError: Upon missing ``based_on`` values or dependency loops.
    """
    if not isinstance(base_pools, OrderedDict):
        base_pools = OrderedDict(list(base_pools))

    plots_cfg = plots_cfg if plots_cfg else {}
    for pcfg_name, pcfg in plots_cfg.items():
        plots_cfg[pcfg_name] = _resolve_based_on(
            pcfg,
            base_pools=OrderedDict(
                tuple(base_pools.items()) + ((label, plots_cfg),)
            ),
            _visited=[(label, pcfg_name)],
        )

    return plots_cfg


def resolve_based_on_single(
    *,
    name: str,
    based_on: Union[str, Sequence[str]],
    plot_cfg: dict,
    **resolve_based_on_kwargs,
) -> dict:
    """Wrapper for :py:func:`~dantro.plot._cfg.resolve_based_on` for cases of
    single independent plot configurations.

    Args:
        name (str): The name of the single plot
        based_on (Union[str, Sequence[str]]): The *extracted* ``based_on``
            argument.
        plot_cfg (dict): The rest of the single plot's configuration. This
            may not include ``based_on``! If this argument evaluates to False,
            will silently assume an empty plots configuration.
        **resolve_based_on_kwargs: Passed on

    """
    return resolve_based_on(
        {name: dict(based_on=based_on, **(plot_cfg if plot_cfg else {}))},
        **resolve_based_on_kwargs,
    )[name]



"""Implements data operations that work on array-like data, e.g. from numpy
or xarray."""

import logging
from typing import Any, Callable, Dict, Iterable, List, Sequence, Tuple, Union

import numpy as np

from .._import_tools import LazyLoader
from ._base_ops import BOOLEAN_OPERATORS

log = logging.getLogger(__name__)

xr = LazyLoader("xarray")

# -----------------------------------------------------------------------------


def apply_along_axis(
    func: Callable, axis: int, arr: np.ndarray, *args, **kwargs
) -> np.ndarray:
    """This is like numpy's function of the same name, but does not try to
    cast the results of func to an :py:class:`numpy.ndarray` but tries to keep
    them as dtype object. Thus, the return value of this function will always
    have one fewer dimension then the input array.

    This goes along the equivalent formulation of
    :py:func:`numpy.apply_along_axis`, outlined in their documentation of the
    function.

    Args:
        func (Callable): The function to apply along the axis
        axis (int): Which axis to apply it to
        arr (numpy.ndarray): The array-like data to apply the function to
        *args: Passed to ``func``
        **kwargs: Passed to ``func``

    Returns:
        numpy.ndarray: with ``func`` applied along ``axis``, reducing the array
            dimensions by one.
    """
    # Get the shapes of the outer and inner iteration; both are tuples!
    shape_outer, shape_inner = arr.shape[:axis], arr.shape[axis + 1 :]
    num_outer = len(shape_outer)

    # These together give the shape of the output array
    out = np.zeros(shape_outer + shape_inner, dtype="object")
    out.fill(None)

    log.debug("Applying function '%s' along axis ...", func.__name__)
    log.debug("  input array:     %s, %s", arr.shape, arr.dtype)
    log.debug("  axis to reduce:  %d", axis)
    log.debug("  output will be:  %s, %s", out.shape, out.dtype)

    # Now loop over the output array and at each position fill it with the
    # result of the function call.
    it = np.nditer(out, flags=("refs_ok", "multi_index"))
    for _ in it:
        midx = it.multi_index

        # Build selector, which has the ellipsis at position `axis`, thus one
        # dimension higher than the out array and matching the input `arr`.
        sel = tuple(midx[:num_outer]) + (Ellipsis,) + tuple(midx[num_outer:])
        log.debug("  midx: %s  -->  selector: %s", midx, sel)

        # Apply function to selected parts of array, then write to the current
        # point in the iteration over the output array.
        out[midx] = func(arr[sel], *args, **kwargs)

    return out


def create_mask(
    data: "xarray.DataArray", operator_name: str, rhs_value: float
) -> "xarray.DataArray":
    """Given the data, returns a binary mask by applying the following
    comparison: ``data <operator> rhs value``.

    Args:
        data (xarray.DataArray): The data to apply the comparison to. This is
            the left-hand-side of the comparison.
        operator_name (str): The name of the binary operator function as
            registered in the ``BOOLEAN_OPERATORS`` database.
        rhs_value (float): The right-hand-side value

    Raises:
        KeyError: On invalid operator name

    Returns:
        xarray.DataArray: Boolean mask
    """
    # Get the operator function
    try:
        comp_func = BOOLEAN_OPERATORS[operator_name]

    except KeyError as err:
        raise KeyError(
            f"No boolean operator '{operator_name}' available! Available "
            f"operators: {', '.join(BOOLEAN_OPERATORS.keys())}"
        ) from err

    # Apply the comparison
    data = comp_func(data, rhs_value)

    # If a name already exists, add some information about the masking to it
    if data.name:
        name = data.name + f" (masked by '{operator_name} {rhs_value}')"
    else:
        name = None

    # Build a new xr.DataArray from that data, retaining all information
    return xr.DataArray(
        data=data, name=name, dims=data.dims, coords=data.coords
    )


def where(
    data: "xarray.DataArray", operator_name: str, rhs_value: float, **kwargs
) -> "xarray.DataArray":
    """Filter elements from the given data according to a condition. Only
    those elemens where the condition is fulfilled are not masked.

    .. note::

        This typically leads to a dtype change to :py:attr:`numpy.float64`.

    Args:
        data (xarray.DataArray): The data to mask
        operator_name (str): The ``operator`` argument used in
            :py:func:`.arr_ops.create_mask`
        rhs_value (float): The ``rhs_value`` argument used in
            :py:func:`.arr_ops.create_mask`
        **kwargs: Passed on to ``.where()`` method call
    """
    return data.where(
        create_mask(data, operator_name=operator_name, rhs_value=rhs_value),
        **kwargs,
    )


def count_unique(data, dims: List[str] = None) -> "xarray.DataArray":
    """Applies :py:func:`numpy.unique` to the given data and constructs a
    :py:class:`xarray.DataArray` for the results.

    NaN values are filtered out.

    Args:
        data: The data
        dims (List[str], optional): The dimensions along which to apply
            np.unique. The other dimensions will be available after the
            operation. If not provided it is applied along all dims.

    """

    def _count_unique(data) -> "xarray.DataArray":
        unique, counts = np.unique(data, return_counts=True)

        # remove np.nan values
        # NOTE np.nan != np.nan, hence np.nan will count 1 for every
        #      occurrence, but duplicate values are not allowed in coords...
        counts = counts[~np.isnan(unique)]
        unique = unique[~np.isnan(unique)]

        if isinstance(data, xr.DataArray) and data.name:
            name = data.name + " (unique counts)"
        else:
            name = "unique counts"

        # Construct a new data array and return
        return xr.DataArray(
            data=counts,
            name=name,
            dims=("unique",),
            coords=dict(unique=unique),
        )

    if not dims:
        return _count_unique(data)

    if not isinstance(data, xr.DataArray):
        raise TypeError(
            f"Data needs to be of type xr.DataArray, but was {type(data)}!"
        )

    # use split-apply-combine along those dimensions not in dims
    split_dims = [dim for dim in data.dims if dim not in dims]

    if len(split_dims) == 0:
        return _count_unique(data)

    data = data.stack(_stack_cu=split_dims).groupby("_stack_cu")
    return data.map(_count_unique).unstack("_stack_cu")


# .............................................................................
# Working with multidimensional data, mostly xarray-based


def populate_ndarray(
    objs: Iterable,
    shape: Tuple[int] = None,
    dtype: Union[str, type, np.dtype] = float,
    order: str = "C",
    out: np.ndarray = None,
    ufunc: Callable = None,
) -> np.ndarray:
    """Populates an empty :py:class:`numpy.ndarray` of the given ``dtype`` with
    the given objects by zipping over a new array of the given ``shape`` and
    the sequence of objects.

    Args:
        objs (Iterable): The objects to add to the :py:class:`numpy.ndarray`.
            These objects are added in the order they are given here. Note
            that their final position inside the resulting array is
            furthermore determined by the ``order`` argument.
        shape (Tuple[int], optional): The shape of the new array. **Required**
            if no ``out`` array is given.
        dtype (Union[str, type, numpy.dtype], optional): dtype of the new
            array. Ignored if ``out`` is given.
        order (str, optional): Order of the new array, determines iteration
            order. Ignored if ``out`` is given.
        out (numpy.ndarray, optional): If given, populates this array rather
            than an empty array.
        ufunc (Callable, optional): If given, applies this unary function to
            each element before storing it in the to-be-returned ndarray.

    Returns:
        numpy.ndarray: The populated ``out`` array or the newly created one (if
            ``out`` was not given)

    Raises:
        TypeError: On missing ``shape`` argument if ``out`` is not given
        ValueError: If the number of given objects did not match the array size
    """
    if out is None and shape is None:
        raise TypeError(
            "Without an output array given, the `shape` argument "
            "needs to be specified!"
        )

    ufunc = ufunc if ufunc is not None else lambda e: e
    out = out if out is not None else np.empty(shape, dtype=dtype, order=order)

    if len(objs) != out.size:
        raise ValueError(
            f"Mismatch between array size ({out.size}, shape: {out.shape}) "
            f"and number of given objects ({len(objs)})!"
        )

    it = np.nditer(out, flags=("multi_index", "refs_ok"))
    for obj, _ in zip(objs, it):
        out[it.multi_index] = ufunc(obj)

    return out


def build_object_array(
    objs: Union[Dict, Sequence],
    *,
    dims: Tuple[str] = ("label",),
    fillna: Any = None,
) -> "xarray.DataArray":
    """Creates a *simple* labelled multidimensional object array.

    It accepts simple iterable types like dictionaries or lists and unpacks
    them into the array, using the key or index (respectively) as coordinate
    for the entry. For dict-like entries, multi-dimensional coordinates can be
    specified by using tuples for keys.
    Subsequently, list-like iterable types (list, tuple etc.) will result in
    one-dimensional output array.

    .. warning::

        This data operation is built for *flexibility*, not for speed. It will
        call the :py:func:`.merge` operation for *every*
        element in the ``objs`` iterable, thus being slow and potentially
        creating an array with many empty elements.
        To efficiently populate an n-dimensional object array, use the
        :py:func:`.populate_ndarray` operation instead
        and build a labelled array from that output.

    Args:
        objs (Union[Dict, Sequence]): The objects to populate the object array
            with. If dict-like, keys are assumed to encode coordinates, which
            can be of the form ``coord0`` or ``(coord0, coord1, …)``, where the
            tuple-form requires as many coordinates as there are entries in the
            ``dims`` argument.
            If list- or tuple-like (more exactly: if missing the ``items``
            attribute) trivial indexing is used and ``dims`` needs to be 1D.
        dims (Tuple[str], optional): The names of the dimensions of the
            labelled array.
        fillna (Any, optional): The fill value for entries that are not
            covered by the dimensions specified by ``objs``. Note that this
            will replace all *null* values, which includes `NaN` but also
            ``None``. This operation is only called if ``fillna is not None``.

    Raises:
        ValueError: If coordinates and/or ``dims`` argument for individual
            entries did not match.
    """

    def get_coords(k, dims: Tuple[str]) -> dict:
        """Turn the iteration key into a valid coordinate dict"""
        if isinstance(k, tuple):
            if len(k) != len(dims):
                raise ValueError(
                    f"Given coordinates {k} could not be matched to the "
                    f"specified dimensions, {dims}! Make sure their sizes "
                    "agree."
                )
            return {_k: [_v] for _k, _v in zip(dims, k)}

        elif len(dims) != 1:
            raise ValueError(
                f"Got scalar coordinate '{k}' but have {len(dims)} dimensions "
                f"({dims}) specified. Either provide an appropriately sized "
                "coordinate tuple or reduce the number of `dims` to one."
            )
        return {dims[0]: [k]}

    # Determine the object iterator
    if not hasattr(objs, "items"):
        if len(dims) != 1:
            raise ValueError(
                "Can only create one-dimensional output data from the given "
                f"list-like object container, but got `dims`: {dims}. "
                f"Instead of {type(objs).__name__}, use a dict to specify "
                "multiple coordinates or adjust the `dims` argument to a "
                "single dimension name."
            )

        it = enumerate(objs)

    else:
        it = objs.items()

    # The (zero-sized) target array
    ndim = len(dims)
    out = xr.DataArray(
        np.zeros((0,) * ndim, dtype="object"),
        dims=dims,
        name="tmp",
        coords=dict(zip(dims, [[]] * ndim)),
    )

    # Populate it entry by entry, merging every entry into the existing array
    for k, v in it:
        coords = get_coords(k, dims)
        new_item = xr.DataArray(
            populate_ndarray([v], shape=(1,) * ndim, dtype="object"),
            dims=dims,
            coords=coords,
            name="tmp",
        )
        out = merge([out, new_item], reduce_to_array=True)

    out.name = None
    if fillna is not None:
        out = out.fillna(fillna)
    return out


def multi_concat(
    arrs: np.ndarray, *, dims: Sequence[str]
) -> "xarray.DataArray":
    """Concatenates :py:class:`xarray.Dataset` or :py:class:`xarray.DataArray`
    objects using :py:func:`xarray.concat`. This function expects the xarray
    objects to be pre-aligned inside the numpy *object* array ``arrs``, with
    the number of dimensions matching the number of concatenation operations
    desired.
    The position inside the array carries information on where the objects that
    are to be concatenated are placed inside the higher dimensional coordinate
    system.

    Through multiple concatenation, the dimensionality of the contained objects
    is increased by ``dims``, while their dtype can be maintained.

    For the sequential application of :py:func:`xarray.concat` along the outer
    dimensions, the custom :py:func:`.apply_along_axis` function is used.

    Args:
        arrs (numpy.ndarray): The array containing xarray objects which are to
            be concatenated. Each array dimension should correspond to one of
            the given ``dims``. For each of the dimensions, the
            :py:func:`xarray.concat` operation is applied along the axis,
            effectively reducing the dimensionality of ``arrs`` to a scalar
            and increasing the dimensionality of the contained xarray objects
            until they additionally contain the dimensions specified in
            the ``dims`` argument.
        dims (Sequence[str]): A sequence of dimension names that is assumed to
            match the dimension names of the array. During each concatenation
            operation, the name is passed along to :py:func:`xarray.concat`
            where it is used to select the dimension of the *content* of
            ``arrs`` along which concatenation should occur.

    Raises:
        ValueError: If number of dimension names does not match the number of
            data dimensions.
    """
    if len(dims) != arrs.ndim:
        raise ValueError(
            f"The given sequence of dimension names, {dims}, did not match "
            f"the number of dimensions of data of shape {arrs.shape}!"
        )

    # Reverse-iterate over dimensions and concatenate them
    for dim_idx, dim_name in reversed(list(enumerate(dims))):
        log.debug(
            "Concatenating along axis '%s' (idx: %d) ...", dim_name, dim_idx
        )

        arrs = apply_along_axis(
            xr.concat, axis=dim_idx, arr=arrs, dim=dim_name
        )
        # NOTE ``np.apply_along_axis`` would be what is desired here, but that
        #      function unfortunately tries to cast objects to np.arrays which
        #      is not what we want here at all! Thus, this function uses the
        #      custom dantro function of the same name instead.

    # Should be scalar now, get the element.
    return arrs.item()


def merge(
    arrs: Union[
        Sequence[Union["xarray.DataArray", "xarray.Dataset"]], np.ndarray
    ],
    *,
    reduce_to_array: bool = False,
    **merge_kwargs,
) -> Union["xarray.Dataset", "xarray.DataArray"]:
    """Merges the given sequence of xarray objects into an
    :py:class:`xarray.Dataset`.

    As a convenience, this also allows passing a :py:class:`numpy.ndarray` of
    dtype ``object`` containing the xarray objects.
    Furthermore, if the resulting :py:class:`xarray.Dataset` contains only a
    single data variable, that variable can be extracted as a
    :py:class:`xarray.DataArray` by setting the ``reduce_to_array`` flag,
    making that array the return value of this operation.
    """
    if isinstance(arrs, np.ndarray):
        arrs = arrs.flat

    dset = xr.merge(arrs, **merge_kwargs)

    if not reduce_to_array:
        return dset

    if len(dset.data_vars) != 1:
        raise ValueError(
            "The Dataset resulting from the xr.merge operation can only be "
            "reduced to a DataArray, if one and only one data variable is "
            "present in the Dataset! "
            f"However, the merged Dataset contains {len(dset.data_vars)} data "
            f"variables:  {', '.join(dset.data_vars)}\n"
            f"Full dataset before attempting to reduce to an array:\n{dset}\n"
            "A typical reason for this is missing data; check that there were "
            "sufficiently populated xarray objects available for merging.\n"
        )

    # Get the name of the single data variable and then get the DataArray
    darr = dset[list(dset.data_vars.keys())[0]]
    # NOTE This is something else than the Dataset.to_array() method, which
    #      includes the name of the data variable as another coordinate. This
    #      is not desired, because it is not relevant.
    return darr


def expand_dims(
    d: Union[np.ndarray, "xarray.DataArray"], *, dim: dict = None, **kwargs
) -> "xarray.DataArray":
    """Expands the dimensions of the given object.

    If the object does not support a ``expand_dims`` method call, it will be
    attempted to convert it to an :py:class:`xarray.DataArray` first.

    Args:
        d (Union[numpy.ndarray, xarray.DataArray]): The object to expand the
            dimensions of
        dim (dict, optional): Keys specify the dimensions to expand, values can
            either be an integer specifying the length of the dimension, or a
            sequence of coordinates.
        **kwargs: Passed on to the ``expand_dims`` method call. For an xarray
            objects that would be :py:meth:`xarray.DataArray.expand_dims`.

    Returns:
        xarray.DataArray: The input data with expanded dimensions.
    """
    if not hasattr(d, "expand_dims"):
        d = xr.DataArray(d)
    return d.expand_dims(dim, **kwargs)


def expand_object_array(
    d: "xarray.DataArray",
    *,
    shape: Sequence[int] = None,
    astype: Union[str, type, np.dtype] = None,
    dims: Sequence[str] = None,
    coords: Union[dict, str] = "trivial",
    combination_method: str = "concat",
    allow_reshaping_failure: bool = False,
    **combination_kwargs,
) -> "xarray.DataArray":
    """Expands a labelled object-array that contains array-like objects into a
    higher-dimensional labelled array.

    ``d`` is expected to be an array *of arrays*, i.e. each element of the
    outer array is an object that itself is an :py:class:`numpy.ndarray`-like
    object. The ``shape`` is the expected shape of each of these *inner*
    arrays. *Importantly*, all these arrays need to have the exact same shape!

    Typically, e.g. when loading data from HDF5 files, the inner array will
    not be labelled but will consist of simple :py:class:`numpy.ndarray`
    objects.
    The arguments ``dims`` and ``coords`` are used to label the *inner* arrays.

    This uses :py:func:`.multi_concat` for concatenating
    or :py:func:`.merge` for merging the object arrays
    into a higher-dimensional array, where the latter option allows for missing
    values.

    .. TODO::

        Make reshaping and labelling optional if the inner array already is a
        labelled array. In such cases, the coordinate assignment is already
        done and all information for combination is already available.

    Args:
        d (xarray.DataArray): The labelled object-array containing further
            arrays as elements (which are assumed to be unlabelled).
        shape (Sequence[int], optional): Shape of the inner arrays. If not
            given, the first element is used to determine the shape.
        astype (Union[str, type, numpy.dtype], optional): All inner arrays
            need to have the same dtype. If this argument is given, the arrays
            will be coerced to this dtype. For numeric data, ``float`` is
            typically a good fallback.
            Note that with ``combination_method == "merge"``, the choice here
            might not be respected.
        dims (Sequence[str], optional): Dimension names for labelling the
            inner arrays. This is necessary for proper alignment. The number of
            dimensions need to match the ``shape``. If not given, will use
            ``inner_dim_0`` and so on.
        coords (Union[dict, str], optional): Coordinates of the inner arrays.
            These are necessary to align the inner arrays with each other. With
            ``coords = "trivial"``, trivial coordinates will be assigned to all
            dimensions. If specifying a dict and giving ``"trivial"`` as value,
            that dimension will be assigned trivial coordinates.
        combination_method (str, optional): The combination method to use to
            combine the object array. For ``concat``, will use dantro's
            :py:func:`.multi_concat`, which preserves
            dtype but does not allow missing values. For ``merge``, will use
            :py:func:`.merge`, which allows missing
            values (masked using ``np.nan``) but leads to the dtype decaying
            to float.
        allow_reshaping_failure (bool, optional): If true, the expansion is not
            stopped if reshaping to ``shape`` fails for an element. This will
            lead to missing values at the respective coordinates and the
            ``combination_method`` will automatically be changed to ``merge``.
        **combination_kwargs: Passed on to the selected combination function,
            :py:func:`.multi_concat` or
            :py:func:`.merge`.

    Returns:
        xarray.DataArray: A new, higher-dimensional labelled array.

    Raises:
        TypeError: If no ``shape`` can be extracted from the first element in
            the input data ``d``
        ValueError: On bad argument values for ``dims``, ``shape``, ``coords``
            or ``combination_method``.
    """

    def prepare_item(
        d: "xarray.DataArray",
        *,
        midx: Sequence[int],
        shape: Sequence[int],
        astype: Union[str, type, np.dtype, None],
        name: str,
        dims: Sequence[str],
        generate_coords: Callable,
    ) -> Union["xarray.DataArray", None]:
        """Extracts the desired element and reshapes and labels it accordingly.
        If any of this fails, returns ``None``.
        """
        elem = d[midx]

        try:
            item = elem.item().reshape(shape)

        except Exception as exc:
            if allow_reshaping_failure:
                return None
            raise ValueError(
                f"Failed reshaping item at {midx} to {shape}! Make sure the "
                f"element\n\n{elem}\n\nallows reshaping. To discard values "
                "where reshaping fails, enable `allow_reshaping_failure`."
            ) from exc

        if astype is not None:
            item = item.astype(astype)

        return xr.DataArray(
            item, name=name, dims=dims, coords=generate_coords(elem)
        )

    # Make sure we are operating on labelled data
    d = xr.DataArray(d)

    # Try to deduce missing arguments and make sure arguments are ok
    if shape is None:
        try:
            shape = d.data.flat[0].shape
        except Exception as exc:
            raise TypeError(
                "Failed extracting a shape from the first element of the "
                f"given array:\n{d}\nCheck that the given array contains "
                "further np.ndarray-like objects. Alternatively, explicitly "
                "provide the `shape` argument."
            ) from exc

    if dims is None:
        dims = tuple(f"inner_dim_{n:d}" for n, _ in enumerate(shape))

    if len(dims) != len(shape):
        raise ValueError(
            "Number of dimension names and number of dimensions of the inner "
            f"arrays needs to match! Got dimension names {dims} for array "
            f"elements of expected shape {shape}."
        )

    if coords == "trivial":
        coords = {n: "trivial" for n in dims}

    elif not isinstance(coords, dict):
        raise TypeError(
            f"Argument `coords` needs to be a dict or str, but was {coords}!"
        )

    if set(coords.keys()) != set(dims):
        raise ValueError(
            "Mismatch between dimension names and coordinate keys! Make sure "
            "there are coordinates specified for each dimension of the inner "
            f"arrays, {dims}! Got:\n{coords}"
        )

    # Handle trivial coordinates for each coordinate dimension separately
    coords = {
        n: (range(l) if isinstance(c, str) and c == "trivial" else c)
        for (n, c), l in zip(coords.items(), shape)
    }

    # Assemble info needed to bring individual array items into proper form
    item_name = d.name if d.name else "data"
    item_shape = tuple(1 for _ in d.shape) + tuple(shape)
    item_dims = d.dims + tuple(dims)
    item_coords = lambda e: dict(
        **{n: [c.item()] for n, c in e.coords.items()}, **coords
    )

    # The array that gathers all to-be-combined object arrays
    arrs = np.empty_like(d, dtype=object)
    arrs.fill(dict())  # are ignored in xr.merge

    # Transform each element to a labelled xr.DataArray that includes the outer
    # dimensions and coordinates; the latter is crucial for alignment.
    # Alongside, type coercion can be performed. For failed reshaping, the
    # element may be skipped.
    it = np.nditer(arrs.data, flags=("multi_index", "refs_ok"))
    for _ in it:
        item = prepare_item(
            d,
            midx=it.multi_index,
            shape=item_shape,
            astype=astype,
            name=item_name,
            dims=item_dims,
            generate_coords=item_coords,
        )

        if item is None:
            # Missing value; need to fall back to combination via merge
            combination_method = "merge"
            continue
        arrs[it.multi_index] = item

    # Now, combine
    if combination_method == "concat":
        return multi_concat(arrs, dims=d.dims, **combination_kwargs)

    elif combination_method == "merge":
        return merge(arrs, reduce_to_array=True, **combination_kwargs)

    raise ValueError(
        f"Invalid combination method '{combination_method}'! "
        "Choose from: 'concat', 'merge'."
    )


# .. Coordinate transformations ...............................................


def transform_coords(
    d: "xarray.DataArray",
    dim: Union[str, Sequence[str]],
    func: Callable,
    *,
    func_kwargs: dict = None,
) -> "xarray.DataArray":
    """Assigns new, transformed coordinates to a data array by applying a
    function on the existing coordinates.

    Uses :py:meth:`xarray.DataArray.assign_coords` to set the new coordinates,
    which returns a shallow copy of the given object.

    Args:
        d (xarray.DataArray): The array to transform the ``dim`` coordinates of
        dim (Union[str, Sequence[str]]): The name or names of the coordinate
            dimension(s) to apply ``func`` to.
        func (Callable): The callable to apply to ``d.coords[dim]``
        func_kwargs (dict, optional): Passed to the function invocation like
            ``func(d.coords[dim], **func_kwargs)``
    """
    if isinstance(dim, str):
        dim = [dim]

    kws = func_kwargs if func_kwargs else {}
    return d.assign_coords({_dim: func(d.coords[_dim], **kws) for _dim in dim})



"""This module provides coordinate parsing capabilities."""

import logging
from itertools import product
from typing import (
    Dict,
    Hashable,
    Iterable,
    List,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

import numpy as np

from ..abc import AbstractDataContainer
from ..tools import is_iterable, try_conversion
from .link import Link, StrongLink

log = logging.getLogger(__name__)

# Some type definitions .......................................................
TDims = Tuple[str]
"""A dimension sequence type"""

TCoord = TypeVar("TCoord", int, float, str, Hashable)
"""A single coordinate value type"""

TCoords = Sequence[TCoord]
"""A sequence of coordinate values"""

TCoordsDict = Dict[str, TCoords]
"""Several coordinates, bundled into a map of dimension name to coordinates"""


# Dimension name extraction ---------------------------------------------------


def extract_dim_names(
    attrs: dict, *, ndim: int, attr_name: str, attr_prefix: str
) -> TDims:
    """Extract dimension names from the given attributes.

    This can be done in two ways:
        1. A list of dimension names was specified in an attribute with the
           name specified by the ``attr_name`` argument
        2. One by one via attributes that start with the string prefix defined
           in ``attr_prefix``. This can be used if not all dimension names are
           available. Note that this will also *not* be used if option 1 is
           used!

    Args:
        attrs (dict): The dict-like object to read attributes from
        obj_logstr (str): A string that is given as context in log and error
            messages, ideally describing the object these attributes belong to
        ndim (int): The expected rank of the dimension names
        attr_name (str): The key to look for in ``attrs`` that would give a
            sequence of the dimension names.
        attr_prefix (str): The prefix to look for in the keys of the ``attrs``
            that would specify the name of a single dimension.

    Returns:
        Tuple[Union[str, None]]: The dimension names or None as placeholder

    Raises:
        TypeError: Attribute found at ``attr_name`` was a string, was not
            iterable or was not a sequence of strings
        ValueError: Length mismatch of attribute found at ``attr_name``
            and the data.

    """
    # The dimension names sequence to populate. None's are placeholders which
    # denote that no information was found for this dimension.
    dim_names = [None] * ndim

    # Distinguish two cases: All dimension names given directly as a list
    # or they are given separately, one by one
    if attr_name in attrs:
        dims_attr = attrs[attr_name]

        # Make sure it is an iterable of strings and of the right length
        if isinstance(dims_attr, str):
            raise TypeError(
                f"Attribute '{attr_name}' needs to be a sequence of "
                f"strings, but not directly a string! Got: {repr(dims_attr)}"
            )

        elif not is_iterable(dims_attr):
            raise TypeError(
                f"Attribute '{attr_name}' needs to be an iterable, but was "
                f"{type(dims_attr)} with value '{dims_attr}'!"
            )

        elif len(dims_attr) != ndim:
            raise ValueError(
                "Number of given dimension names does not match "
                f"the given rank of {ndim}! Names given: {dims_attr}"
            )

        # Data seems ok.
        # Create the sequence of dimension name, potentially needing to do
        # some further processing ...
        for dim_num, dim_name in enumerate(dims_attr):
            # Might be a numpy scalar or 1-sized array; resolve that
            if isinstance(dim_name, np.ndarray):
                dim_name = dim_name.item()

            if not isinstance(dim_name, str):
                raise TypeError(
                    "Dimension names need to be strings, got "
                    f"{type(dim_name)} with value '{dim_name}' for "
                    f"dimension {dim_num}!"
                )

            dim_names[dim_num] = dim_name

    # Have a populated list of dimension names now, all strings.
    # There might be additionally specified dimension names that overwrite
    # the names given here...

    # Go over all attributes and check for the prefix
    for attr_name, attr_val in attrs.items():
        if attr_name.startswith(attr_prefix):
            # Extract the integer dimension number
            try:
                dim_num = int(attr_name[len(attr_prefix) :])

            except ValueError as err:
                raise ValueError(
                    "Could not extract the dimension number from "
                    f"the container/group attribute named '{attr_name}'! "
                    "Take care that the part after the prefix "
                    f"('{attr_prefix}') can be converted to an integer."
                ) from err

            # Make sure its valid
            if dim_num < 0 or dim_num >= ndim:
                raise ValueError(
                    f"The dimension number {dim_num:d} extracted from "
                    f"attribute '{attr_name}' exceeds the given rank {ndim}!"
                )

            # Make sure the attribute value is a string
            if isinstance(attr_val, np.ndarray):
                attr_val = attr_val.item()  # ... assuming already decoded.

            if not isinstance(attr_val, str):
                raise TypeError(
                    "Dimension names need be strings, but the "
                    f"attribute '{attr_name}' provided {type(attr_val)} "
                    f"with value '{attr_val}'!"
                )

            # All good now. Write it to the dim name list
            dim_names[dim_num] = attr_val

    # Done.
    return tuple(dim_names)


# Coordinate extraction -------------------------------------------------------
# Low level extractor functions ...............................................
# These should all take the coordinate arguments as the first argument and then
# accept arbitrary keyword-only arguments, of which some may be used to
# determine the coordinate values


def _coords_start_and_step(
    cargs, *, data_shape: tuple, dim_num: int, **__
) -> Iterable[int]:
    """Interpret as integer start and step of range expression and use the
    length of the data dimension as number of steps
    """
    start, step = cargs
    stop = start + (step * data_shape[dim_num])
    return range(int(start), int(stop), int(step))


def _coords_trivial(
    _, *, data_shape: tuple, dim_num: int, **__
) -> Iterable[int]:
    """Returns trivial coordinates for the given dimension by creating a
    range iterator from the selected data shape.
    """
    return range(data_shape[dim_num])


def _coords_scalar(coord, **__) -> List[TCoord]:
    """Returns a single, scalar coordinate, i.e.: list of length 1"""
    if isinstance(coord, np.ndarray):
        coord = coord.item()

    # If it is not iterable, it is a scalar now. Expected return value is a
    # list, though, so convert it.
    if not is_iterable(coord):
        coord = [coord]

    if len(coord) != 1:
        raise ValueError(f"Expected scalar coordinate, but got: {coord}!")

    # Make sure it is a list, not a tuple
    return list(coord)


def _coords_linked(cargs, *, link_anchor_obj, **__) -> Link:
    """Creates a Link object which is to be used for coordinates"""
    # Need to parse potential numpy array arguments to string
    if isinstance(cargs, np.ndarray):
        cargs = cargs.item()

    # Problem: at this point, the container to be linked to might not
    # know its full path within the data tree. Thus, coordinate
    # resolution has to be postponed until it is clear.
    # For that reason, create a link object, which can forward to an
    # actual container once the coordinates are applied...
    return StrongLink(anchor=link_anchor_obj, rel_path=cargs)


# Map of extractors ...........................................................
# fmt: off
COORD_EXTRACTORS = {
    "values":           lambda cargs, **__: cargs,
    "range":            lambda cargs, **__: range(*cargs),
    "arange":           lambda cargs, **__: np.arange(*cargs),
    "linspace":         lambda cargs, **__: np.linspace(*cargs),
    "logspace":         lambda cargs, **__: np.logspace(*cargs),
    "start_and_step":   _coords_start_and_step,
    "trivial":          _coords_trivial,
    "scalar":           _coords_scalar,
    "linked":           _coords_linked,
}
# fmt: on

# Actual (high-level) extraction functions ....................................


def extract_coords_from_attrs(
    obj: Union[AbstractDataContainer, np.ndarray],
    *,
    dims: Tuple[Union[str, None]],
    strict: bool,
    coords_attr_prefix: str,
    default_mode: str,
    mode_attr_prefix: str = None,
    attrs: dict = None,
) -> TCoordsDict:
    """Extract coordinates from the given object's attributes.

    This is done by iterating over the given ``dims`` and then looking
    for attributes that are prefixed with ``coords_attr_prefix`` and ending in
    the name of the dimension, e.g. attributes like ``coords__time``.

    The value of that attribute is then evaluated according to a so-called
    attribute ``mode``. By default, the mode set by ``default_mode`` is used,
    but it can be set explicitly for each dimension by the ``mode_attr_prefix``
    parameter.

    The resulting number of coordinates for a dimension always need to match
    the length of that dimension. However, the corresponding error can only be
    raised once this information is applied.

    Args:
        obj (Union[AbstractDataContainer, numpy.ndarray]): The object to
            retrieve the attributes from (via the ``attrs`` attribute). If the
            ``attrs`` *argument* is given, will use those instead.
            It is furthermore expected that this object specifies the shape of
            the numerical data the coordinates are to be generated for by
            providing a ``shape`` property. This is possible with
            :py:class:`~dantro.containers.numeric.NumpyDataContainer` and
            derived classes.
        dims (Tuple[Union[str, None]]): Sequence of dimension names; this
            may also contain None's, which are ignored for coordinates.
        strict (bool): Whether to use strict checking, where no additional
            coordinate-specifying attributes are allowed.
        coords_attr_prefix (str): The attribute name prefix for coordinate
            specifications
        default_mode (str): The default coordinate extraction mode. Available
            modes:

            * ``values``: the explicit values (iterable) to use for coordinates
            * ``range``: range arguments
            * ``arange``: np.arange arguments
            * ``linspace``: np.linspace arguments
            * ``logspace``: np.logspace arguments
            * ``trivial``: The trivial indices. This does not require a value
              for the coordinate argument.
            * ``scalar``: makes sure only a single coordinate is provided
            * ``start_and_step``: the start and step values of an integer range
              expression; the stop value is deduced by looking at the length of
              the corresponding dimension. This is then passed to the python
              range function as (start, stop, step)
            * ``linked``: Load the coordinates from a linked object within the
              tree; this works only if ``link_anchor_obj`` is part of a data
              tree at the point of coordinate resolution!

        mode_attr_prefix (str, optional): The attribute name prefix that can
            be used to specify a non-default extraction mode. If not given, the
            default mode will be used.
        attrs (dict, optional): If given, these attributes will be used instead
            of attempting to extract attributes from ``obj``.

    Returns:
        TCoordsDict: The ``(dim_name -> coords)`` mapping

    Raises:
        ValueError: On invalid coordinates mode or (with strict attribute
            checking) on superfluous coordinate-setting attributes.

    """

    def get_coord(attrs: dict, dim_name: str, dim_num: int):
        """Determines coordinate values for a single dimension."""
        # The argument the coordinate will be determined from; may be None.
        cargs = attrs.get(coords_attr_prefix + dim_name)

        # Determine the mode to interpret the attribute values
        mode = default_mode
        if mode_attr_prefix:
            mode = attrs.get(mode_attr_prefix + dim_name, default_mode)

        # Might have to process the mode, e.g. because it is an array-type
        if isinstance(mode, np.ndarray):
            mode = mode.item()

        # Check if the mode is available
        if mode not in COORD_EXTRACTORS:
            _mode_attr_name = mode_attr_prefix + dim_name
            _extraction_modes = ", ".join(COORD_EXTRACTORS.keys())
            raise ValueError(
                f"Invalid mode '{mode}' to interpret coordinate "
                "attribute values! Check whether a mode "
                f"attribute '{_mode_attr_name}' is set. "
                f"Available modes: {_extraction_modes}"
            )

        # Invoke the method, passing on available arguments
        try:
            return COORD_EXTRACTORS[mode](
                cargs,
                dim_num=dim_num,
                data_shape=obj.shape,
                link_anchor_obj=obj,
            )

        except Exception as exc:
            raise type(exc)(
                "Failed extracting coordinates for dimension "
                f"'{dim_name}' via extraction mode '{mode}'! {exc}"
            ) from exc

    # If necessary, get the attributes from the given object
    if not attrs:
        attrs = obj.attrs

    # The to-be-populated mapping from dimension names to coordinates.
    coords_map = dict()

    # Get the coordinates for all labelled (!) dimensions; if not labelled, no
    # coordinate is expected.
    for dim_num, dim_name in enumerate(dims):
        if dim_name is None:
            continue

        # Is labelled; try to extract a coordinate and store it if found
        coords = get_coord(attrs, dim_name, dim_num)
        if coords is not None:
            coords_map[dim_name] = coords

    # Done.
    # Without strict attribute checking, can already return here
    if not strict:
        return coords_map

    # Need to do strict attribute checking ...
    # Determine the prefixes that are to be checked
    prefixes = [coords_attr_prefix]
    if mode_attr_prefix:
        prefixes.append(mode_attr_prefix)

    # Check all attribute name and prefix combinations
    for attr_name, prefix in product(attrs.keys(), prefixes):
        # See whether there are matching attributes that were not already
        # extracted above, i.e. appear as keys in the map
        if (
            attr_name.startswith(prefix)
            and attr_name[len(prefix) :] not in coords_map
        ):

            _dims_avail = ", ".join([d for d in dims if d is not None])
            _prefixes = ", ".join(prefixes)

            raise ValueError(
                f"Got superfluous attribute '{attr_name}' that "
                "does not match any of the available names of "
                f"labelled dimensions: {_dims_avail}. Valid "
                f"attribute prefixes: {_prefixes}. Either remove "
                "the attribute or turn strict attribute checking "
                "off."
            )

    # All good. Return the coordinate map.
    return coords_map


def extract_coords_from_name(
    obj: AbstractDataContainer,
    *,
    dims: TDims,
    separator: str,
    attempt_conversion: bool = True,
) -> TCoordsDict:
    """Given a container or group, extract the coordinates from its name.

    The name of the object may be a ``separator``-separated string, where each
    segment contains the coordinate value for one dimension.

    This function assumes that the coordinates for each dimension are scalar.
    Thus, the values of the returned dict are sequences of length 1.

    Args:
        obj (AbstractDataContainer): The object to get the coordinates of by
            inspecting its name.
        dims (TDims): The dimension names corresponding to the coordinates that
            are expected to be found in the object's name.
        separator (str): The separtor to apply on the name.
        attempt_conversion (bool, optional): Whether to attempt conversion of
            the string value to a numerical type.

    Returns:
        TCoordsDict: The coordinate dict, i.e. a mapping from the external
            dimension names to the coordinate values. In this case, there can
            only a single value for each dimension!

    Raises:
        ValueError: Raised upon failure to extract external coordinates:
            On ``ext_dims`` evaluating to False, f coordinates were missing for
            any of the external dimensions, if the number of coordinates
            extracted from the name did not match the number of external
            dimensions, if any of the strings extracted from the object's name
            were empty.
    """
    # Split the string and make some basic checks.
    coords = obj.name.split(separator)

    if len(coords) != len(dims):
        raise ValueError(
            "Number of coordinates extracted from the name of "
            f"{obj.logstr} does not match the number of expected "
            f"dimensions! Parsed coordinates: {coords}. "
            f"Dimensions: {dims}"
        )

    if not all(coords):
        raise ValueError(
            "One or more of the coordinates extracted from the "
            f"name of {obj.logstr} were empty! Got: {coords}"
        )

    # Build the dict, attempting conversion of the objects.
    coords = {
        dim_name: [try_conversion(c_val) if attempt_conversion else c_val]
        for dim_name, c_val in zip(dims, coords)
    }

    return coords


def extract_coords_from_data(
    obj: AbstractDataContainer, *, dims: TDims
) -> TCoordsDict:
    """Tries to extract the coordinates from the data of the given container
    or group. For that purpose, the ``obj`` needs to support the ``coords``
    property.

    Args:
        obj (AbstractDataContainer): The object that holds the data from which
            the coordinates are to be extracted.
        dims (TDims): The sequence of dimension names for which the coordinates
            are to be extracted.
    """
    return {dim_name: list(obj.coords[dim_name].values) for dim_name in dims}


# .............................................................................


def extract_coords(
    obj: AbstractDataContainer,
    *,
    mode: str,
    dims: TDims,
    use_cache: bool = False,
    cache_prefix: str = "__coords_cache_",
    **kwargs,
) -> TCoordsDict:
    """Wrapper around the more specific coordinate extraction functions.

    .. note::

        This function does not support the extraction of non-dimension
        coordinates.

    Args:
        obj (AbstractDataContainer): The object from which to extract the
            coordinates.
        mode (str): Which mode to use for extraction. Can be:

            * ``name``:   Use the name of the object
            * ``attrs``:  Use the attributes of the object
            * ``data``:   Use the data of the object

        dims (TDims): The dimensions for which the attributes are to be
            extracted. All dimension names given here are expected to be found.
        use_cache (bool, optional): Whether to use the object's attributes to
            write an extracted value to the cache and read it, if available.
        cache_prefix (str, optional): The prefix to use for writing the cache
            entries to the object attributes. Will suffix this with ``dims``
            and ``coords`` and store the respective data there.
        **kwargs: Passed on to the actual coordinates extraction method.

    Raises:
        NotImplementedError: If ``use_cache`` is set
    """
    EXTRACTORS = dict(
        name=extract_coords_from_name,
        attrs=extract_coords_from_attrs,
        data=extract_coords_from_data,
    )

    if use_cache:
        # TODO Read from cache
        raise NotImplementedError("use_cache")

    # Get the extraction function
    if mode not in EXTRACTORS:
        _extraction_modes = ", ".join(EXTRACTORS.keys())
        raise ValueError(
            f"Invalid extraction mode '{mode}'! Valid modes: "
            f"{_extraction_modes}"
        )

    extraction_func = EXTRACTORS[mode]
    try:
        coords = extraction_func(obj, dims=dims, **kwargs)

    except Exception as exc:
        raise type(exc)(
            f"Failed extracting coordinates of {obj} using "
            f"mode '{mode}': {exc}"
        ) from exc

    # TODO Write to cache

    return coords



"""Subpackage that organizes general plotting utilities."""

from .color_mngr import ColorManager, parse_cmap_and_norm_kwargs
from .mpl import figure_leak_prevention
from .plot_func import PlotFuncResolver, is_plot_func
